import os
import sys
import time
import numpy as np
import re
from collections import OrderedDict

from mmgroup.mat24 import vect_to_cocode
from mmgroup.mat24 import ploop_theta
from mmgroup.mm import mm_aux_index_sparse_to_leech2
from mmgroup.mm import mm_vector
from mmgroup.mm import mm_aux_mmv_extract_sparse_signs
from mmgroup.mm import mm_aux_mmv_extract_sparse
from mmgroup.structures.mm0_group import MM0
from mmgroup.mm_space import MMV, MMVector, MMSpace
from mmgroup.generators import mm_group_check_word_n
from mmgroup.generators import mm_group_words_equ
from mmgroup.generators import mm_group_n_mul_element
from mmgroup.generators import mm_group_n_reduce_element 
from mmgroup.generators import gen_leech3to2_type4
from mmgroup.generators import gen_leech2_reduce_type4
from mmgroup.clifford12 import chk_qstate12
from mmgroup.clifford12 import uint64_parity
from mmgroup.clifford12 import leech2matrix_add_eqn
from mmgroup.clifford12 import leech2matrix_solve_eqn
from mmgroup.clifford12 import bitmatrix64_t
from mmgroup.clifford12 import xsp2co1_half_order_word
from mmgroup.clifford12 import xsp2co1_power_word
from mmgroup.mm15 import op_eval_A_rank_mod3 as mm_op15_eval_A_rank_mod3 

from mmgroup.mm15 import op_copy as mm_op15_copy
from mmgroup.mm15 import op_compare as mm_op15_compare
from mmgroup.mm15 import op_word as mm_op15_word
from mmgroup.mm15 import op_norm_A as mm_op15_norm_A
from mmgroup.mm15 import op_watermark_A as mm_op15_watermark_A

from mmgroup.dev.mm_reduce import find_order_vector
from mmgroup.dev.mm_reduce.find_order_vector import stabilizer_vector
from mmgroup.dev.mm_reduce.find_order_vector import find_vector_71_mod3
from mmgroup.dev.mm_reduce.find_order_vector import find_vector_v94_mod5


MMV3 = MMV(3)
MMV15 = MMV(15)
MM = MM0  #  TODO: Fixme

ORDER_VECTOR = None
ORDER_TAGS = None


OFS_NORM_A = 0
OFS_DIAG_VA = 1
OFS_WATERMARK_PERM = 2
OFS_TAGS_Y = 26
OFS_SOLVE_Y = 37
OFS_TAGS_X = 48
OFS_SOLVE_X = 72
OFS_TAG_SIGN = 96


_DIR = os.path.split(find_order_vector.__file__)[0]
PY_FILENAME = os.path.join(_DIR, "order_vector_data.py")




#######################################################################
# Parse string to vector or group element
#######################################################################

m_vect = re.compile("MV<([0-9]+)")

def parse_str(value):
    if isinstance(value, str):
        mo = m_vect.match(value)
        if mo:
            p = int(mo.groups()[0])
            return MMVector(p, value)
        elif value.startswith("M0"):
            return MM0(value)
    return s



#######################################################################
# Assemble a test vector mod 15 from the input data
#######################################################################


def mm_element_from_data(data):
    if isinstance(data, str):
        result = parse_str(data)
        assert type(result) == MM0
        return result
    elif isinstance(data, MM0):
        return MM0(data)
    return MM0('a', data)

def vector_from_data(factor, data):
    if isinstance(data, str):
        v = parse_str(data)
    elif isinstance(data, MMVector):
        v = data
    else:
        v = MMV15('S', data)
    assert type(v) == MMVector, type(v)
    return MMV15(factor, v)

def identity_vector(factor):
    return factor * MMV15("+".join(["D_%s" % i for i in range(24)]))


   


def make_order_vector(s_g71, s_v71, s_gA, diag, s_g94, s_v94):
    v71 = vector_from_data(10, s_v71)
    g71 = mm_element_from_data(s_g71)
    w71 = stabilizer_vector(v71, g71, 71)
    assert w71 is not None
    w71 *= mm_element_from_data(s_gA)
    v94 = vector_from_data(6, s_v94)
    g94 = mm_element_from_data(s_g94)
    w94 = stabilizer_vector(v94 - v94 * g94, g94**2, 47)
    assert w94 is not None
    if not isinstance(diag, int): diag = diag[0]
    w = w71 + w94 + identity_vector(5 * diag % 15)
    diag = 0
    v3 = mm_op15_eval_A_rank_mod3(w.data, diag) & 0xffffffffffff
    assert v3 != 0
    v_type4 = gen_leech3to2_type4(v3)
    assert v_type4 == 0x800000
    w.reduce()
    return w



def map_y(y_index):
    i, j = (y_index >> 14) & 0x1f, (y_index >> 8) & 0x1f
    vect = (1 << i) + (1 << j)
    gc = vect_to_cocode(vect)
    assert 0 <= gc < 0x800
    return gc 
    
    
def map_x(x_index):
    v2 = mm_aux_index_sparse_to_leech2(x_index) 
    return ((v2 & 0xfff) << 12) | ((v2 >> 12) & 0xfff)    


def make_order_tags(order_vector, tags_y, tags_x, tag_sign):
    ov = order_vector.data
    tags_y = np.array(tags_y, dtype = np.uint32) 
    tags_x = np.array(tags_x, dtype = np.uint32)
    tag_sign =  np.array(tag_sign, dtype = np.uint32) 
    watermark_perm = np.zeros(24, dtype = np.uint32)
    ok = mm_op15_watermark_A(ov, watermark_perm)
    assert ok >= 0

    solve_yt = np.zeros(11, dtype = np.uint64)
    assert len(tags_y) == 11
    nrows = 0
    for y in tags_y:
        eqn = map_y(y)
        nrows += leech2matrix_add_eqn(solve_yt, nrows, 11, eqn)
    assert nrows == 11, nrows
    solve_y = list(bitmatrix64_t(solve_yt, 11))
    assert len(solve_y) == 11
    assert mm_aux_mmv_extract_sparse_signs(15, ov, tags_y, 11) == 0
    
    solve_xt = np.zeros(24, dtype = np.uint64)
    assert len(tags_x) == 24
    nrows = 0
    for i, x in enumerate(tags_x):
        eqn =  map_x(x)  
        nrows += leech2matrix_add_eqn(solve_xt, nrows, 24, eqn)
    assert nrows == 24, nrows
    solve_x = list(bitmatrix64_t(solve_xt, 24))
    
    # Concatenate computed lists to the global numpy array 'ORDER_TAGS'
    order_tags = np.array(sum(map(list, [
        [mm_op15_norm_A(ov), 0], 
        watermark_perm, tags_y, solve_y, tags_x, solve_x, tag_sign
    ]), []), dtype = np.uint32)
    assert len(order_tags) == 97, len(order_tags)
    t0 = mm_aux_mmv_extract_sparse_signs(
        15, ov, order_tags[OFS_TAGS_X:], 24)
    assert t0 == 0
    return order_tags



def order_vector_from_dict(d):
    S_G71, S_V71 = d["S_G71"], d["S_V71"]
    S_GA, DIAG_VA = d["S_GA"], d["DIAG_VA"]
    S_G94, S_V94 = d["S_G94"], d["S_V94"]
    data1 = [S_G71, S_V71, S_GA, DIAG_VA, S_G94, S_V94]
    ov = make_order_vector(*data1)
    TAGS_Y, TAGS_X =  d["TAGS_Y"], d["TAGS_X"]
    TAG_SIGN = d["TAG_SIGN"]
    data2 = [TAGS_Y, TAGS_X, TAG_SIGN]
    o_tag = make_order_tags(ov, *data2)
    keys =  [
      "S_G71", "S_V71", "S_GA", "DIAG_VA", "S_G94", "S_V94",
      "TAGS_Y", "TAGS_X", "TAG_SIGN"
    ]
    d1 = {k: d[k] for k in keys}
    return ov, o_tag, d1



#######################################################################
# Check that the test vector supports reduction
#######################################################################

Y_INDICES = [("A", i, j) for i in range(2) for j in range(i+1, 24)]
X_INDICES = [("B", i, j) for i in range(2) for j in range(i+1, 24)]
X_INDICES += [("C", 0, j)  for j in range(1, 24)]
BASIS = [0] + [1 << i for i in range(11)]
X_INDICES += [("X", i, j) for j in range(0, 24)  for i in  BASIS]
del BASIS


    
def eqn_system(vector, tag_indices, map_f, n):
    entries = [vector[index] for index in tag_indices]
    indices = [MMSpace.index_to_sparse(*x) for x in tag_indices]
    matrix= np.zeros(24, dtype = np.uint64)
    rows, cols = 0, n
    out_indices = []
    for (entry, index) in zip(entries, indices):
        if 1 <= entry < 15:
            eqn = map_f(index)
            new_rows = leech2matrix_add_eqn(matrix, rows, cols, eqn)
            if new_rows:
                out_indices.append(index)
                rows += new_rows
                if rows == n:
                    break
    if rows < n:
        return None
    mask = (1 << n) - 1
    out_matrix = [int(matrix[i]) & mask for i in range(n)] 
    return  out_indices   


def eqn_sign(vector):
    for j in range(24):
        if 1 <= vector["Z", 0, j] < 15:
            return [ MMSpace.index_to_sparse("Z", 0, j) ]
    return None


def augment_v_data(v, data):
    a = np.array(data, dtype = np.uint32)
    mm_aux_mmv_extract_sparse(15, v.data, a, len(a))
    return [x for x in a]


def check_v(v, verbose = 0):
    vB = v['B']
    for p in (3,5):
        if (vB % p == 0).all():
            if verbose:
                print("Vector may be zero (mod %d)" % p)
            return None
    mark = np.zeros(24, dtype = np.uint32)
    if mm_op15_watermark_A(v.data, mark) < 0:
        if verbose:
            print("Permutation watermarking failed")
        return None
    result_y = eqn_system(v, Y_INDICES, map_y, 11)
    if result_y is None:
        if verbose:
            print("Check Y failed")
        return result
    result_x = eqn_system(v, X_INDICES, map_x, 24)
    if result_x is None:
        if verbose:
            print("Check X failed")
        return result
    result_sign = eqn_sign(v)
    if result_sign is None:
        if verbose:
            print("Check for sign failed")
        return result
    results = [result_y, result_x, result_sign]
    return tuple([augment_v_data(v, data) for data in results])
    



#######################################################################
# Seach for the relevant data
#######################################################################

m_vect = re.compile("MV<([0-9]+)")

def to_list(value):
    if isinstance(value, str):
        value = parse_str(value)
    if isinstance(value, MM0):
        return list(map(int, value.mmdata))
    if isinstance(value, MMVector):
        return list(map(int, value.as_sparse()))
    if isinstance(value, int):
        return [value]
    if isinstance(value, list):
        return list(map(int, value))
    err = "Cannot convert type-'%s' object to list"
    raise TypeError(err % type(value))



def str_data(text, data):
    data = to_list(data)
    s = "%s = [\n   " % text
    for i, x in enumerate(data):
        s += hex(x) + ","
        s += "\n   " if i % 6 == 5 else " "
    s += "\n]\n"
    return s


def find_order_vector(verbose = 1):
    verbose = 1
    if verbose:
        print("Trying to find a vector of order 71")
    s_g71, s_v71, s_gA, diag = find_vector_71_mod3(verbose)
    if verbose:
        print("Trying to find a vector of order 94")
    for trials in range(200,-1,-1):
        s_g94, s_v94 = find_vector_v94_mod5(verbose = verbose)
        if s_v94 is None:
            continue
        v = make_order_vector(s_g71, s_v71, s_gA, diag, s_g94, s_v94)
        tag_data = check_v(v, verbose=verbose)
        if tag_data is not None:
            if verbose:
                print("Tests for v94 passed")
            break
    if not trials:
        err = "No suitable vector in the monster representation found"
        raise ValueError(err) 
    v_data =  s_g71, s_v71, s_gA, diag, s_g94, s_v94
    V_NAMES =  ["S_G71", "S_V71", "S_GA", "DIAG_VA", "S_G94", "S_V94"]
    TAG_NAMES =  [ "TAGS_Y", "TAGS_X", "TAG_SIGN"]   
    if verbose:        
        for text, data in zip(TAG_NAMES, tag_data):
            print(str_data(text, data))
    result = OrderedDict(zip(V_NAMES + TAG_NAMES, v_data + tag_data))
    return result


HEADER = """# This file has been created automatically, do not change!
# For documentation see module mmgroup.dev.mm_reduce.order_vector.

"""

def write_order_vector(result):
    print("Writing file " + PY_FILENAME)
    f = open(PY_FILENAME, "wt")
    print(HEADER, file = f)
    for text, data in result.items():
        print(str_data(text, data), file = f)
    f.close()
    



def get_order_vector(recompute = False, verbose = 0):
    try:
        from mmgroup.dev.mm_reduce import order_vector_data
        assert not recompute
    except (ImportError, AssertionError):
        result = find_order_vector(verbose)
        write_order_vector(result)
        from mmgroup.dev.mm_reduce import order_vector_data
    ov, o_tag, data = order_vector_from_dict(order_vector_data.__dict__)
    return ov, o_tag, data









"""Create a temporary directory for tests

For some tests we want to create a temporary directory in a
known position in the source file tree, so that C programs or 
python extensions can be created end executed in that temporary 
directory.

Function make_tmp_dir(parent) creates such a temporary directory 
and returns its path name. Here 'parent' can be a path name or a 
python module object. Function kill_tmp_dir(parent) deletes that 
directory. Function temp_dir_name(parent) returns the path name
of the temporary directory without creating it. 

If '<parent>' is a python module object, then '<parent>.temp' 
is a module object referring to the temporary directory created. 
"""

import os
import shutil
import time
import types

TMPNAME = "temp"

temp_dir = None

def temp_dir_name(parent= None):
    """Return absolute path name of temporary directory

    Here argument 'parent' is as in function make_temp_dir().
    This function does not create any directory.
    """
    if isinstance(parent, types.ModuleType):
        dir_ = os.path.dirname(os.path.realpath(parent.__file__))
        return  os.path.abspath(os.path.join(dir_, TMPNAME))
    elif parent:
        return os.path.abspath(os.path.join(parent, TMPNAME))
    else:
        return os.path.abspath(TMPNAME)


def make_temp_dir(parent = None): 
    """Create a temporary directory, return its absolute path name

    Optionally, a 'parent' directory of the temporary directory
    may be given. By default, this is the current directory.

    Alternatively, 'parent' be a module object which has been 
    imported with an "import" statement. Then a directory 'temp' 
    is created as a subdirectory of the directory of that module, 
    which is os.path.dirname(os.path.realpath(module.__file__)).
    Futhermore, and empty file with name "__init__.py" is created
    in that subdirectory 'temp'. Then a python module foo.py
    written into that subdirectory can be imported as
    <parent>.temp.foo, with <parent> the module object 'parent'.

    Caution: this destroys an existing file or directory with name
    "temp" in the parent directory.
    """ 
    global temp_dir  
    temp_dir =  temp_dir_name(parent)
    try:
        shutil.rmtree(temp_dir)
        time.sleep(0.1) # It take a while to delete TMPNAME
    except FileNotFoundError:
        pass
    os.mkdir(temp_dir)
    if isinstance(parent, types.ModuleType):
        f = open(os.path.join(temp_dir, "__init__.py"), "wt")
        f.write("\n")
        f.close()
    return temp_dir


def kill_temp_dir(name):
    """Remove a temporary directory

    Here 'name' must be the path name returned from the last
    recent call to function make_tmp_dir().
    """
    global temp_dir  
    if name and name == temp_dir:
        try:
            shutil.rmtree(temp_dir)
        except FileNotFoundError:
            pass
        temp_dir = None
    


    


"""Wrapper for module mm_group_n.c for the subgroup N_0 of the monster

 TODO: adjust documentation to new version

 That group N_0 is described in [Conw85] and [Seys19]. More precisely, 
 we actually calculate in the fourfold cover N of N_0 of structure
  
      2**2 . 2**2 . 2**11 . 2**(11+11) . (Sym3 \times Mat24),

 using the notation in [Seys19]. Here Sym3 is the symmetric permutation
 group of 3 elements and Mat24 is the Mathieu group acting on 24 elements. 
 An element of N is represented as a 5-tuple (exp, f, e, delta, pi) of 
 integers of type uint32_t. Here


 exp    represents the element t**exp for the triality element t
        (t is called 'tau' in [Seys19], section 5);
 f      represents the element y_f, 0 <= f < 0x2000 an element of Pl;   
 e      represents the element x_e, 0 <= e < 0x2000 an element of Pl;   
 delta  represents the element x_delta, where 0 <= delta < 0x1000 is
        an element of the Golay cocode C*;
 pi     represents the element (pi) of Mat24 with number pi.  
        More precisely, pi represents the preimage of (pi) in the 
        automorphism group AutPl of the Parker loop Pl that maps all 
        basis vectors of Pl to positive elements of Pl.
        
 Elements tau, x_f, x_e, x_delta are defined as in [Seys19], setion 5. 
 For the numbering of the elements of the Parker loop Pl, the Golay
 cocode C*, and the Mathieu group Mat24, see module mat24_functions.c.
 The basis of the Parker loop Pl is also given in that module.
 For the notion of a positive element of the Parker loop (with 
 respect to a certain cocycle) of the Parker loop see [Seys19]. The 
 cocycle selected for Pl is also given in mat24_functions.c.
 The elements of N are multiplied in the order given by the 5-tuple.

 An element of the group N is represented as an instance of class 
 FastGroupN_Word.

 For interfacing with class FastGroupN_Word, generating elements
 of the group N may be given as tuples as follows:

 ('t', exp)     means  t**exp for an integer exp.
 ('y', f)       means  y_f for an integer f representing an
                element of the Parker loop.
 ('x', e)       means  x_e for an integer e representing an
                element of the Parker loop.
 ('d', delta)   means x_delta for an integer delta representing
                an element of the Golay cocode 
 ('p', pi)      means the element (pi) defined as follows:

The element (pi) may be given as a list of 24 integers representing
an element of Mat24 as described in module mat24_functions.c. It
may also be given as an integer representing the element of Mat24 
with that number, as described in module mat24_functions.c.

An instance of class FastGroupN_Word representing a element g of N
may be right multiplied with another element g1 of N using the '*'
operator. Then g1 may be:

   - a tuple representing an element of N as described above,
   - a list of such tuples (not yet implemented),
   - another instance of class FastGroupN_Word,
   - an element of group N, where N is an instance of class 
     mgroup_n.MGroupN representing a group similar to N. 
   
Note that class mgroup_n.MGroupN has methods for generating 
elements of the free product <N, T> of N and a group L of order 3 
generated by an element l. There is a homomorphsim from <N, L>
to the monster MM which maps N to the subgroup N_0 of MM 
discussed in [Conw85] and [Seys19] and l to the element xi
of MM defined in [Seys19], section 9.4.

For the sake of precision we mention that class mgroup_n.MGroupN 
may also represent some factor groups N' of N such that N_0 is a
factor group of N'. 

The tuples given above are also supported by the methods of class
mgroup_n.MGroupN. In addition, the tuple ('l', exp) represents the
element l**exp of MM. So we can describe elements of the monster MM 
by sequences of such tuples.

Class FastGroupN_Word executres the group operations in the group N 
by using the fast routines in module mm_group_n.c.

Expnonentiation is also supported in the usual way, with g**2 = g*g  
etc., g**(-1) the inverse of g, and g1**g2 = g2**(-1) * g1 * g2.
""" 
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import re
import numpy as np
from numbers import Integral
from random import randint
import warnings
import time



from mmgroup.structures.abstract_mm_group import AbstractMMGroupWord
from mmgroup.structures.abstract_mm_group import AbstractMMGroup
from mmgroup.structures.parse_atoms import  AtomDict      
from mmgroup.structures.parse_atoms import ihex       
from mmgroup.structures.construct_mm import iter_mm       
from mmgroup.structures.construct_mm import load_group_name     


from mmgroup.generators import mm_group_n_mul_atom
from mmgroup.generators import mm_group_n_mul_element
from mmgroup.generators import mm_group_n_inv_element
from mmgroup.generators import mm_group_n_mul_inv_element
from mmgroup.generators import mm_group_n_mul_delta_pi, mm_group_n_mul_x
from mmgroup.generators import mm_group_n_mul_inv_delta_pi
from mmgroup.generators import mm_group_n_mul_y, mm_group_n_mul_t
from mmgroup.generators import mm_group_n_reduce_element
from mmgroup.generators import mm_group_n_copy_element
from mmgroup.generators import mm_group_n_right_coset_N_x0

from mmgroup.mat24 import pow_ploop



######################################################################
# Modelling an element of the group
######################################################################


class GroupN_Word(AbstractMMGroupWord):
    def __init__(self,  tag = None, atom = None, *args, **kwds):
        self.data = np.zeros(5, dtype = np.uint32)
        atoms = iter_mm(self.group, tag, atom)
        for a in atoms:
            mm_group_n_mul_atom(self.data, a)
        mm_group_n_reduce_element(self.data)

    ####################################################################
    # Alternative multipication and division methods for tests
    ####################################################################

    def mul(self, g):
        return self.group.mul(self, g)
    def mul_atoms(self, g):
        return self.group.mul_atoms(self, g)
    def div(self, g):
        return self.group.div(self, g)
    def div_atoms(self, g):
        return self.group.div_atoms(self, g)

    @property
    def mmdata(self):
        element = self.data
        mm_group_n_reduce_element(element)
        a = []
        if element[0]: a.append(0x50000000 + element[0])
        if element[1]: a.append(0x40000000 + element[1])
        if element[2]: a.append(0x30000000 + element[2])
        if element[3]: a.append(0x10000000 + element[3])
        if element[4]: a.append(0x20000000 + element[4])
        return np.array(a, dtype = np.uint32)

    def right_coset_N_x0(self):
        """Decmpose element as right coset of ``N_x0``

        The function returns two group elements ``g_Nx0``, ``g_t``
        such that the product of these two elements is equal to
        this element, ``g_Nx0`` is in the subgroup ``N_x0`` of
        ``N_0``, and  ``g_t`` is a power of the triality element.
        """
        g_Nx0 = GroupN_Word()
        g_t = GroupN_Word()
        mm_group_n_copy_element(self.data, g_Nx0.data)
        g_t.data[0] = mm_group_n_right_coset_N_x0(g_Nx0.data)
        return g_Nx0, g_t
        


######################################################################
# Modelling the group
######################################################################

class GroupN(AbstractMMGroup):
    __instance = None
    group_name = "GroupN"
    word_type = GroupN_Word  # type of an element (=word) in the group 
    __instance = None

    def __new__(cls):
        if GroupN.__instance is None:
             GroupN.__instance = AbstractMMGroup.__new__(cls)
        return GroupN.__instance

    def __init__(self):
        super(GroupN, self).__init__()
        

    def copy_word(self, g1):
        """Return deep copy of group element g1"""
        g_copy = self.word_type(group=self)
        g_copy.data[:] = g1.data
        return g_copy

    def _equal_words(self, g1, g2):
        """Return True iff elements g1 and g2 are equal 

        This method is called for elements g1 and g2 of the group
        'self' only.
		
        In concrete group this method should be overwritten with
        a comparison of the relevant attributes of g1 and g2.
        """
        mm_group_n_reduce_element(g1.data)
        mm_group_n_reduce_element(g2.data)        
        return (g1.data == g2.data).all()


   
    def __call__(self, tag = None, atom = None):
       return self.word_type(tag, atom) 

    ####################################################################
    # group operation
    ####################################################################

        
    def _imul(self, g1, g2):
        mm_group_n_mul_element(g1.data, g2.data, g1.data)
        return g1

    def _invert(self, g1):
        g2 = GroupN_Word(group=self)
        mm_group_n_inv_element(g1.data, g2.data)
        return g2

    ####################################################################
    # alternative multiplication and division methods for tests
    ####################################################################

    def mul(self, g1, g2):
        assert g2.group == self
        mm_group_n_mul_element(g1.data, g2.data, g1.data)
        return g1

    def mul_atoms(self, g1, g2):
        assert g2.group == self
        for tag, data in g2.as_tuples():
           self.mul_atom(g1, tag, data)
        return g1

    def div(self, g1, g2):
        assert g2.group == self
        mm_group_n_mul_inv_element(g1.data, g2.data, g1.data)
        return g1

    def div_atoms(self, g1, g2):
        assert g2.group == self
        for tag, data in reversed(g2.as_tuples()):
           self.divide_atom(g1, tag, data)
        return g1

    ####################################################################
    # Low-level group operation
    ####################################################################


    def mul_atom(self, g, tag, data):
        """Auxiliary method for method mul()        """
        v = g.data
        if tag == "d":
            mm_group_n_mul_delta_pi(v, data, 0)
        elif tag == "p":
            mm_group_n_mul_delta_pi(v, 0, data)
        elif tag == 'x':
            mm_group_n_mul_x(v, data)
        elif tag == 'y':
            mm_group_n_mul_y(v, data)
        elif tag == 't':
            mm_group_n_mul_t(v, data)
        else:
            raise TypeError("Illegal tag %s in group word" % t)

    def divide_atom(self, g, tag, data):
        """Auxiliary method for method mul()        """
        v = g.data
        if tag == "d":
            mm_group_n_mul_inv_delta_pi(v, data, 0)
        elif tag == "p":
            mm_group_n_mul_inv_delta_pi(v, 0, data)
        elif tag == 'x':
            mm_group_n_mul_x(v, pow_ploop(data, 3))
        elif tag == 'y':
            mm_group_n_mul_y(v, pow_ploop(data, 3))
        elif tag == 't':
            mm_group_n_mul_t(v, -int(data) % 3)
        else:
            raise TypeError("Illegal tag %s in group word" % t)


StdGroupN = GroupN()
GroupN_Word.group = StdGroupN
load_group_name(StdGroupN)




r"""Generation of  C code dealing with the Mathieu group Mat24

Generating the ``mmgroup.clifford12`` extension
..........................................

Function ``clifford12_make_c_code()`` generates C code for basic 
computations in the Clifford group of up to 12 qubits.

The generated C modules are used in the python extension 
``mmgroup.clifford``. The functions used by that extension are 
contained in a shared library with name ``mmgroup_clifford12.dll``. 
The reason for creating such a shared library is that that these 
functions are also called by C functions written for other python 
extensions.

We use the C code generation mechanism in class 
``generate_c.TableGenerator``. Here a .c file  and a .h file is
crreated from files ``*.ske`` in subdirectory
``src/mmgroup/dev/clifford12``. The .ske file is like a .c file, but 
augmented with some code generation statements for entering tables 
and automatically generated code into the .c file to be generated. 
This .ske file may also have statements for automatically 
generating a .h file declaring the exported functions.  


Location of the output files
............................

The location of the generated output files is controlled by certain
variables in module config.py. Each of these variables specifies the
name of a directory.

Files with extension .c, .h go to the directory ``C_DIR``. Files with
extension .pxd, .pxi, .pyx go to the directory ``PXD_DIR``. 
"""



from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import types
import sys
import re
import os
import subprocess
import shutil



from config import SRC_DIR, DEV_DIR,  C_DIR, PXD_DIR
from config import REAL_SRC_DIR 
sys.path.append(REAL_SRC_DIR)

import mmgroup
from mmgroup.generate_c import TableGenerator, make_doc
from mmgroup.generate_c import pxd_to_pyx



VERBOSE = 0

##########################################################################
# Generating .c files
##########################################################################




SKE_DIR = os.path.join(DEV_DIR, "clifford12")

SIMPLE_SKE_FILES = ["qstate12", "qmatrix12"]
PXI_SKE_FILES = [
    "bitmatrix64",  "uint_sort", "xsp2co1", 
    "leech3matrix", "xsp2co1_elem",
    "involutions", "xsp2co1_traces",
]
SKE_FILES = SIMPLE_SKE_FILES + PXI_SKE_FILES


##########################################################################
# Generating .h and .pxd files files
##########################################################################



H_FILE_NAME = "clifford12.h"      # name of generated .h file


PXD_FILE_NAME = "clifford12.pxd"  # name of generated .pxd file


H_FILE_BEGIN = """

// %%EXPORT_KWD CLIFFORD12_API

// %%GEN h

/** @file clifford12.h
 File ``clifford.h`` is the header file for shared library 
 ``mmgroup_clifford12``. This comprises the C modules

  qstate12.c, qmatrix12.c,
  bitmatrix64.c, uint_sort.c, xsp2co1.c, leech3matrix.c, 
  xsp2co1_elem.c, involutions.c,
*/

#ifndef CLIFFORD12_H
#define CLIFFORD12_H

/// @cond DO_NOT_DOCUMENT 

#include <stdint.h>

#define CLIFFORD12_DLL  // We want a DLL!!


// Generic helper definitions for shared library support
#if defined _WIN32 || defined __CYGWIN__
  #define CLIFFORD12_HELPER_DLL_IMPORT __declspec(dllimport)
  #define CLIFFORD12_HELPER_DLL_EXPORT __declspec(dllexport)
  #define CLIFFORD12_HELPER_DLL_LOCAL
#else
  #if __GNUC__ >= 4
    #define CLIFFORD12_HELPER_DLL_IMPORT __attribute__ ((visibility ("default")))
    #define CLIFFORD12_HELPER_DLL_EXPORT __attribute__ ((visibility ("default")))
    #define CLIFFORD12_HELPER_DLL_LOCAL  __attribute__ ((visibility ("hidden")))
  #else
    #define CLIFFORD12_HELPER_DLL_IMPORT
    #define CLIFFORD12_HELPER_DLL_EXPORT
    #define CLIFFORD12_HELPER_DLL_LOCAL
  #endif
#endif


// Now we use the generic helper definitions above to define CLIFFORD12_API 
// and CLIFFORD12_LOCAL.
// CLIFFORD12_API is used for the public API symbols. It either DLL imports 
// or DLL exports (or does nothing for static build). 
// CLIFFORD12_LOCAL is used for non-api symbols.

#ifdef CLIFFORD12_DLL // defined if CLIFFORD12 is compiled as a DLL
  #ifdef CLIFFORD12_DLL_EXPORTS // defined if we are building the CLIFFORD12 DLL 
                           // (instead of using it)
    #define CLIFFORD12_API CLIFFORD12_HELPER_DLL_EXPORT
  #else
    #define CLIFFORD12_API CLIFFORD12_HELPER_DLL_IMPORT
  #endif // CLIFFORD12_DLL_EXPORTS
  #define CLIFFORD12_LOCAL CLIFFORD12_HELPER_DLL_LOCAL
#else // CLIFFORD12_DLL is not defined: this means CLIFFORD12 is a static lib.
  #define CLIFFORD12_API
  #define CLIFFORD12_LOCAL
#endif // CLIFFORD12_DLL


/// @endcond 

"""


H_FILE_END = """
// %%GEN h

#ifdef __cplusplus
}
#endif
#endif  // #ifndef CLIFFORD12_H
"""



### Declarations for the generated .pxd files


PXD_DECLARATIONS = """

from libc.stdint cimport uint64_t, uint32_t, uint16_t, uint8_t
from libc.stdint cimport int64_t, int32_t, int8_t


cdef extern from "clifford12.h":
    enum: QSTATE12_MAXCOLS
    enum: QSTATE12_MAXROWS
    enum: QSTATE12_UNDEF_ROW
    ctypedef struct qstate12_type:
        uint32_t maxrows
        uint32_t nrows
        uint32_t ncols
        int32_t  factor
        uint32_t shape1
        uint32_t reduced
        uint64_t *data 


""".format()




##########################################################################
# Entering tables and automatically generated code
##########################################################################


TABLE_CLASSES = [
]



##########################################################################
# Generating the .pyx file
##########################################################################


##########################################################################
# Generating the .pyx file
##########################################################################



PXI_FILE_NAME = "clifford12.pxi"



##########################################################################
# Functions of this module for generating .c, .h and .pxi files
##########################################################################



def list_c_files():
    """Return list of names of c files generated by make_clifford12()"""
    def c_file(name):
         return os.path.join(C_DIR, name + ".c")
    return [c_file(name) for name in SKE_FILES]




def make_clifford12():
    """Generate clifford12 .c files from  .ske files 

    The relevant .ske files are listed in variable
    SKE_FILES. One corresponding .h and .pxd file
    is also generated. The .c files and the .h file are written 
    to directory given by C_DIR, the .pxd file is written to the 
    current directory.

    Tables and directives for automatically generated C code
    are taken from the classes listed in BASIC_TABLE_CLASSES.

    Return pair of lists, one of the .c files and one of the.
    .pxd files generated
    """
    tables = {}
    directives = {}
    global generated_tables
    for table_class in TABLE_CLASSES:
        table_instance = table_class()
        tables.update(table_instance.tables)
        directives.update(table_instance.directives)
    # print("Clifford 12 functions:\n",  directives.keys())
    tg = TableGenerator(tables, directives, verbose = VERBOSE)
    # first generate C files
    c_files = []
    all_ske_files = [H_FILE_BEGIN]
    for name in SKE_FILES:
        ske_file = name + ".ske"
        ske_path = os.path.join(SKE_DIR, ske_file)
        c_file = name + ".c"
        c_path = os.path.join(C_DIR, c_file)
        print("Creating %s from %s" % (c_file, ske_file))
        tg.generate(ske_path, c_path)
        all_ske_files.append(ske_path)
        c_files.append(c_path)
    all_ske_files.append (H_FILE_END)

    # generate .h and .pxd file
    h_file =  H_FILE_NAME
    h_path =  os.path.join(C_DIR, h_file)
    pxd_file =  PXD_FILE_NAME
    print("Creating %s from previous .ske files" % h_file)
    tg.generate(all_ske_files, None, h_path)
    tg.generate_pxd(
        os.path.join(PXD_DIR, pxd_file), 
        h_file, 
        PXD_DECLARATIONS
    )
    return c_files,  [ pxd_file ]


##########################################################################
# The main function for generating code
##########################################################################


def generate_files():
    """The main function of this module for generating code

    This function generates all reqired .c files and also the required
    headers.

    It also generates the .pxi file with name given by PYX_FILE_NAME and
    stores it in the current directory. Cython will use the .pxi file to
    build a wrapper for all generated C functions. 
    """
    def pxi_comment(text, f):
        print("\n" + "#"*70 + "\n### %s\n" % text + "#"*70 + "\n\n",file=f)
    c_files,  pxd_files =  make_clifford12() 
    f_pxi = open(os.path.join(PXD_DIR, PXI_FILE_NAME), "wt")
    print(PXD_DECLARATIONS, file = f_pxi)
    
    for pxd_f in pxd_files:
        pxi_comment("Wrappers for C functions from file %s" % pxd_f, f_pxi)
        pxi_content = pxd_to_pyx(
            os.path.join(PXD_DIR, pxd_f),
            os.path.split(pxd_f)[0],
            select = True
        )
        print(pxi_content, file = f_pxi)
     
    f_pxi.close()
    return c_files



##########################################################################
# Main program
##########################################################################


if __name__ == "__main__":
    generate_files()









r"""Generation of genernal C code for reduction in the monster 

This script generates the C code for reduction of elements of the
monster group. The generated C modules are used in the python 
extension ``mmgroup.mm_reduce``. The functions used by that extension 
are contained in a shared library with name ``mmgroup_mm_reduce``. 
The reason for creating such a shared library is that that these 
functions may also be called by an external computer algebra package.

Therefore several .c, .h, .pyd files are created, and also one .pyx
file. Then the Cython package is used for integrating all C 
functions into a single python extension. The action of the extension
is controlled by the .pyx file, which also includes the generated 
.pyd files.

Function ``generate_pyx()`` in this module generates all these files.


Implementation of this module is along the lines of module ``codegen_mm"


"""


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import re
import collections
import warnings
from numbers import Integral


from config import SRC_DIR, DEV_DIR,  C_DIR, PXD_DIR
from config import REAL_SRC_DIR 

from config import INT_BITS
SKE_DIR = os.path.join(DEV_DIR, "mm_reduce")



if  __name__ == "__main__":
    sys.path.append(REAL_SRC_DIR)
    from mmgroup.generate_c import TableGenerator, make_doc, format_item
    from mmgroup.generate_c import pxd_to_pyx
    from mmgroup.dev.mm_op import mm_op
    from mmgroup.dev.mm_reduce import order_vector_tables




VERBOSE = 0


##########################################################################
# Entering tables and automatically generated code
##########################################################################

TABLE_CLASSES = None

def table_classes():
    global TABLE_CLASSES
    if not TABLE_CLASSES is None:
        return  TABLE_CLASSES

    if "mockup" in sys.argv[1:]:
        TABLE_CLASSES = [
            mm_op.Mockup_MM_Op,
            order_vector_tables.Mockup_OrderVectorTable,
        ]
    else:
        TABLE_CLASSES = [
            mm_op.MM_Op,
            order_vector_tables.OrderVectorTable,
        ]  
    return TABLE_CLASSES




##########################################################################
# Generating .c files
##########################################################################


H_REDUCE_SKELETONS = [
    "mm_reduce.h",
]

C_REDUCE_SKELETONS = [
   "mm_order_vector",
   "mm_order",
   "mm_reduce",
   "mm_suborbit",
 
]



def mm_reduce_sources():
    return C_REDUCE_SKELETONS



##########################################################################
# Generating .h and .pxd files files
##########################################################################

H_REDUCE_NAME = "mm_reduce.h"      # name of generated .h file


PXD_REDUCE_NAME = "mm_reduce.pxd"  # name of generated .pxd file



### Generate the basic header


H_REDUCE_BEGIN = """

// %%EXPORT_KWD MM_REDUCE_API


// %%GEN h
#ifndef MM_REDUCE_H
#define MM_REDUCE_H


"""


H_REDUCE_END = """
// %%GEN h

#endif  // #ifndef MM_REDUCE_H
"""




### Declarations for the generated .pxd files


PXD_DECLARATIONS = """

from libc.stdint cimport uint64_t, uint32_t, uint16_t, uint8_t
from libc.stdint cimport int64_t, int32_t
from libc.stdint cimport uint{INT_BITS}_t as uint_mmv_t

INT_BITS = {INT_BITS}
""".format(INT_BITS = INT_BITS)






    
##########################################################################
# Generating the .pyx file
##########################################################################



PXI_FILE_NAME = "mm_reduce.pxi"







##########################################################################
# Functions of this module for generating .c, .h and .pyx file
##########################################################################

def list_reduce_c_files():
    """Return list of names of c files generated by make_reduce()"""
    def c_file(name):
         return os.path.join(C_DIR, name + ".c")
    return [c_file(name) for name in C_REDUCE_SKELETONS]


def make_reduce():
    """Generate basic .c files from  .ske files 

    The relevant .ske files are listed in variable
    C_REDUCE_SKELETONS. One corresponding .h and .pxd file
    is also generated. The .c files and the .h file are written 
    to directory given by C_DIR, the .pxd file is written to the 
    current directory.

    Tables and directives for automatically generated C code
    are taken from the classes listed in BASIC_TABLE_CLASSES.

    Return pair of lists, one of the .c files and one of the.
    .pxd files generated
    """
    tables = {}
    directives = {}
    global generated_tables
    for table_class in table_classes():
        table_instance = table_class(15)
        tables.update(table_instance.tables)
        directives.update(table_instance.directives)
    # print("Basic functions:\n",  directives.keys())
    tg = TableGenerator(tables, directives, verbose = VERBOSE)
    # first generate C files
    c_files = []
    all_ske_files = [H_REDUCE_BEGIN] 
    all_ske_files += [
        os.path.join(SKE_DIR, f) for f in H_REDUCE_SKELETONS]
    for name in C_REDUCE_SKELETONS:
        ske_file = name + ".ske"
        ske_path = os.path.join(SKE_DIR, ske_file)
        c_file = name + ".c"
        c_path = os.path.join(C_DIR, c_file)
        print("Creating %s from %s" % (c_file, ske_file))
        tg.generate(ske_path, c_path)
        all_ske_files.append(ske_path)
        c_files.append(c_path)
    all_ske_files.append(H_REDUCE_END)

    # generate .h and .pxd file
    h_file =  H_REDUCE_NAME
    h_path =  os.path.join(C_DIR, h_file)
    pxd_file =  PXD_REDUCE_NAME
    print("Creating %s from previous .ske files" % h_file)
    #print(all_ske_files)
    tg.generate(all_ske_files, None, h_path)
    print("Creating %s" % pxd_file)
    tg.generate_pxd(
        os.path.join(PXD_DIR, pxd_file), 
        h_file, 
        PXD_DECLARATIONS
    )
    return c_files,  [ pxd_file ]







##########################################################################
# The main function for generating code
##########################################################################


def generate_files():
    """The main function of this module for generating code

    This function generates all reqired .c files and also the required
    headers.

    It also generates the .pxi file with name given by PYX_FILE_NAME and
    stores it in the current directory. Cython will use the .pxi file to
    build a wrapper for all generated C functions. 
    """
    def pxi_comment(text, f):
        print("\n" + "#"*70 + "\n### %s\n" % text + "#"*70 + "\n\n",file=f)
    c_files,  pxd_files =  make_reduce() 
    print("Creating %s" % PXI_FILE_NAME)
    f_pxi = open(os.path.join(PXD_DIR, PXI_FILE_NAME), "wt")
    print(PXD_DECLARATIONS, file = f_pxi)
    for pxd_f in pxd_files:
        pxi_comment("Wrappers for C functions from file %s" % pxd_f, f_pxi)
        pxi_content = pxd_to_pyx(
            os.path.join(PXD_DIR, pxd_f),
            os.path.split(pxd_f)[0]
        )
        print(pxi_content, file = f_pxi)
    f_pxi.close()
    print("Code generation for module mm_reduce terminated successfully")
    return c_files


##########################################################################
# Main function
##########################################################################





if __name__ == "__main__":
    generate_files()



"""We deal with the cocode of the Golay code

Documentation see module mmgroup.structure.gcode
"""
from functools import reduce
from operator import __xor__
from numbers import Integral, Number
from random import randint


try:
    # Try importing the fast C function
    from mmgroup import mat24 
except (ImportError, ModuleNotFoundError):
    # Use the slow python function if the C function is not available
    from mmgroup.dev.mat24.mat24_ref import  Mat24
    mat24 = Mat24


from mmgroup.structures.abstract_group import AbstractGroupWord
from mmgroup.structures.parity import Parity
from mmgroup.structures.parse_atoms import ihex

from mmgroup.structures.gcode import as_vector24, str_vector24
from mmgroup.structures.gcode import GCode, GcVector



ERR_RAND = "Illegal string for constructing type %s element" 

ERR_DIV4 = "%s object may be divided by 2 or 4 only"
ERR_DIV2 = "%s object may be divided by 2 only"


#######################################################################
# Import derived classed
#######################################################################

import_pending = True

def complete_import():
    """Internal function of this module

    If you need any of the objects declared above, do the following:

    if import_pending:
        complete_import()
    """
    global import_pending, PLoop
    global AutPL, AutPlGroup, XLeech2
    from mmgroup.structures.ploop import PLoop
    from mmgroup.structures.autpl import AutPL, AutPlGroup
    from mmgroup.structures.xleech2 import XLeech2
    import_pending = False




#######################################################################
# Class Cocode 
#######################################################################




class Cocode():
    """This class models an element of the cocode of the Golay code.

    The Golay code is a binary ``[24, 12, 8]`` code. So its cocode has 
    :math:`2^{12}` elements, each of it is a :math:`24` bit long word
    (modulo the Golay code).

    The :math:`2^{12}` Golay cocode elements are numbered from
    ``0`` to ``0xfff``. Bit ``i`` in that number refers to the
    coefficient of the  ``i``-th basis vector of the Golay cocode,
    which may be :math:`0` or :math:`1`. The chosen basis of the
    cocode is the reciprocal basis of the chosen basis of the
    Golay code.

    :param value:

      This is the value of the Golay cocode word to be returned. 

    :type value: see table below for legal types

    :return: A Golay cocode element
    :rtype:  an instance of class |Cocode|

    :raise:
        * TypeError if ``type(value)`` is not in the table given below.


    Depending on its type parameter **value** is  interpreted as follows:

    .. table:: Legal types for constructor of class ``Cocode``
      :widths: 20 80

      ==================== ================================================
      type                 Evaluates to
      ==================== ================================================
      ``int``              Here the cocode element with number ``value`` 
                           is returned. ``0 <= value <= 0x1000`` must hold.

      ``list`` of ``int``  A list of integers  ``0 <= i < 24`` is 
                           interpreted as a list of bit positions to be
                           set in a bit vector. The cocode word
                           corresponding to that bit vector is returned.

      class |Cocode|       A deep copy of parameter ``value`` is 
                           returned. 

      class |XLeech2|      The *cocode* part of the |XLeech2| object  
                           is converted to a cocode element and 
                           returned. 

      ``str``              Create random element depending on the string
                              | ``'r'``: Create arbitrary cocode element
                              | ``'e'``: Create an even cocode element
                              | ``'o'``: Create an odd cocode element

      ==================== ================================================


    **Standard operations**
    
    Addition of cocode elements and scalar multiplication (with scalars 
    of  type  ``int`` or |Parity|) are done as usual.

    ``g & c`` and ``c & g`` are legal operations for a Golay code
    element ``g`` of type  |GCode| and a cocode element ``c`` of type  
    |Cocode|. The result of this operation is not well defined as 
    a bit vector, but it  has a well-defined parity. Thus
    ``g & c`` returns a parity object of type |Parity|.

    **Standard functions**

    Let ``c`` be a cocode element of type |Cocode|.
    ``len(c)`` returns the bit weight of a shortest representative
    of the Golay code element ``c``. 

    **Special functions**
  
    The expression ``c % 2`` is a shorthand for  ``Parity(len(c))``.
    It returns the parity of ``c`` as a parity
    object of type  |Parity|. 

    """
    __slots__ = "value"
    basis = mat24.basis[:12]
    parity_class = True

    def __init__(self, value):
        if import_pending:
            complete_import()
        if isinstance(value, Integral):
            self.value = value & 0xfff
        elif isinstance(value, Cocode):
            self.value = value.value
        elif isinstance(value, GcVector):
            self.value = mat24.vect_to_cocode(value.value)
        elif isinstance(value, XLeech2):
            v = value.value
            self.value = (mat24.ploop_theta(v >> 12) ^ v) & 0xfff
        elif isinstance(value, str):
            if len(value) == 1 and value in "reo":
                self.value = randint(0, 0xfff)
                if "o" in value: self.value |= 0x800
                if "e" in value: self.value &= 0x7ff
            else:
                raise ValueError(ERR_RAND % 'Cocode')
        else:
            vector = as_vector24(value)
            self.value = mat24.vect_to_cocode(vector)


    def __eq__(self, other):
        return (isinstance(other, Cocode)  
            and (self.value ^ other.value) & 0xfff == 0)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __pos__(self):
        return self
        
    __neg__ = __pos__

    def __and__(self, other):
        if isinstance(other, GCode):
            return Parity(mat24.scalar_prod(self.value, other.value))
        else:
            return NotImplemented


    def __add__(self, other):
        if isinstance(other, Cocode):
            return Cocode((self.value ^ other.value) & 0xfff)
        elif isinstance(other, GcVector):
            return Cocode(self.value ^ mat24.vect_to_cocode(other.value))
        elif other == 0:
            return Cocode(self)
        else:
            return NotImplemented


    __sub__ = __add__
    __radd__ = __add__
    __rsub__ = __radd__

    def __mul__(self, other):
        if isinstance(other, Integral):
            mask = -(other & 1)
            return Cocode(self.value & mask & 0xfff)
        elif isinstance(other, AutPL):
            return Cocode(mat24.op_cocode_perm(self.value, other.perm))
        else:
            return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, Integral):
            mask = -(other & 1)
            return Cocode(self.value & mask & 0xfff)
        else:           
            return NotImplemented

    def __len__(self):
        return mat24.cocode_weight(self.value)

    def half_weight(self):
        err ="Don't know the halved weight of that cocode vector"
        raise ValueError(err)

    def __truediv__(self, other):
        if isinstance(other, Integral):
            if abs(other) == 1:
                return self
            elif abs(other) == 2:
                return self.half_weight()
            else:
                raise ValueError(ERR_DIV2 % type(self))
        else:
            return NotImplemented

    def __mod__(self, other):
        return Parity(self.parity) % other

    @property
    def ord(self):
        """Return the number of the cocode element.

        We have ``0 <= i < 0x1000`` for the returned number ``i``.
        """
        return self.value & 0xfff
        
    @property
    def cocode(self):
        """Return the number of the cocode element.

        We have ``0 <= i < 0x1000`` for the returned number ``i``.

        Same as property ``ord``.
        """
        return self.value & 0xfff

    @property    
    def parity(self):
        """Return parity of the cocode element as a |Parity| object.

        """
        return (self.value >> 11) & 1


    def syndrome(self, i=None):
        """Return syndrome of cocode element as a bit vector.

        :param i: ``None`` (default) or an integer ``0 <= i < 24`` 
                  used to select a syndrome.
                 
        :return: Syndrome of a Golay cocode element as a bit
                 vector of type |GcVector|.

        :raise:
           *  ValueError if argument ``i`` is ``None`` and the
              syndrome is not unique.

        Any Golay cocode element has either a unique shortest
        representative of bit weight ``<= 3`` or precisely six
        shortest representatives of bit weight ``4`` which form
        a partition of the underlying set of the code. Such a
        partition is called a *tetrad*. 

        In coding theory, a shortest representative of a cocode 
        element is called a *syndrome*.
             
        If the syndrome is unique, the function returns that 
        syndrome. Otherwise it returns the syndrome containing
        bit at position ``i``.
        """
        if i is None: i = 24
        return GcVector(mat24.cocode_syndrome(self.value, i))

    def syndrome_list(self, i=None):
        """Return syndrome of cocode element as list of bit positions.

        :param i: ``None`` (default) or an integer ``0 <= i < 24`` 
                  used to select a syndrome.
                 
        :return: Syndrome of a Golay cocode element as a list
                 of at most four bit positions.

        :raise:
            * ValueError if argument ``i`` is ``None`` and the
              syndrome is not unique.

        The syndrome of the Golay cocode element is calculated
        in the same way as in method **syndrome**. But here
        the result is returned as an ordered list of bit positions
        corresponding to the bits which are set in the syndrome.
        """
        if i is None: i = 24
        syn = mat24.cocode_syndrome(self.value, i)
        return [i for i in range(24) if (syn >> i) & 1]



    def str(self):
        return "<Cocode_%s>" % ihex(self.value & 0xfff, 3)
    __repr__ = str    


    @classmethod
    def str_basis(cls):
       b = "Golay cocode basis c_i[0,...,23], i = 0,...,11"
       return str_basis(b, "c", cls.basis)

    @classmethod
    def show_basis(cls):
        print(cls.str_basis())









#######################################################################
# Subclasses of class Cocode
#######################################################################



class PLoopIntersection(Cocode):
    """Models the intersection of two Golay code words as a cocode word.
   
    Not for public use!
    """
    __slots__ = "value", "v1", "v2"
    parity = 0

    def __init__(self, v1, v2):
        assert isinstance(v1, GCode) and isinstance(v2, GCode) 
        self.v1, self.v2 = v1.value, v2.value
        self.value = mat24.ploop_cap(self.v1, self.v2)

    def half_weight(self):
        return Parity(mat24.ploop_comm(self.v1, self.v2))
    
    def str(self):
        return "<Cocode_%s>" % ihex(self.value & 0xfff, 3)
    __repr__ = str    




 


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



import sys
import warnings
from random import sample, randint
import re


from copy import deepcopy
from collections.abc import Sequence


from mmgroup.structures.abstract_group import AbstractGroupWord
from mmgroup.structures.abstract_group import AbstractGroup





####################################################################
####################################################################
### Class AutoGroup and helpers for that class
####################################################################
####################################################################



####################################################################
### Exceptions provided for the user
####################################################################


"""Exception for a user-defined rewrite rule

The function implementing a rewrite rule may detect that it cannot
improve the input word given to it as an argument.

In this case it must throw the following excption AutoGroupMulError
to indicate that the application of that rule has failed.

A typical example is the case where a rule is applied to all 
elements of a subgroup G to simplify these elements. Then it
will try to simplify any element e of G and it will return the 
empty word if e is the neutral element of G. But if e cannot be
further simplfied, we must return an indication of failure to 
prevent infinite reapplication of the same rule.
""" 
class AutoGroupMulError(Exception):
    """Multiplication of group elements is not possible"""
    pass
    
####################################################################
### Class ReverseTaggedRewriteNode and helpers for that class
####################################################################



class ReverseTaggedRewriteNode(object):
    def __init__(self, parent = None, tag = None):
        self.children = {}
        self.action = None
        self.parent = parent
        if not parent is None:
            assert not tag in self.parent.children
            parent.children[tag] = self

    def child(self, tag):
        try:
            return self.children[tag]
        except KeyError:
            return ReverseTaggedRewriteNode(self, tag)

    def set_action(self, action):
        if not self.action is None:
            raise ValueError("Duplicate rewrite action")
        self.action = action



class TaggedRewriteSystem(object):
    def __init__(self, rules):
        self.word = None
        self.rules = ReverseTaggedRewriteNode()
        self.translation_rules = {}
        for tags, rule in rules.items():
            self.add_rule(tags, rule)

    def add_rule(self, tags, rule):
        if len(tags) == 1:
            assert not tags[0] in self.translation_rules
            self.translation_rules[tags[0]] = rule
        else:
            node = self.rules
            for tag in reversed(tags):
                node = node.child(tag)
            node.set_action(rule)

    def didactic_subst_word(self, group, word):
        """

        Didactic version: starts with shortest possible rule, and
        orders priority of rules by ascending length.
        """
        node = self.rules
        for i, w in enumerate(reversed(word)):
            try:
                node = node.children[w.tag]
            except KeyError:
                return 0, None
            if node.action:
                try:
                    return i + 1, node.action(group, word[-i - 1:])
                except AutoGroupMulError:
                    continue
        return 0, None
                     
    def subst_word(self, group, word):
        """

        Pragmatic version: starts with rule of length 2, then it tries 
        rules of length 3, 4, 5,... , and finally, it tries rule of
        length 1 if present.

        Dealing too much with rules of length 1 is a bit like self
        satifsction and causes too much overhead.
        """
        node = self.rules
        for i, w in enumerate(reversed(word)):
            try:
                node = node.children[w.tag]
            except KeyError:
                break
            if node.action:
                try:
                    return i + 1, node.action(group, word[-i - 1:])
                except AutoGroupMulError:
                    continue
        if len(word) and word[-1].tag in self.translation_rules:
            rule = self.translation_rules[word[-1].tag]
            try:
                return 1, rule(group, word[- 1:])
            except AutoGroupMulError:
                return 0, None
        return 0, None




####################################################################
####################################################################
### Class AutoGroup and helpers for that class
####################################################################
####################################################################






####################################################################
### Class AutoGroupWord
####################################################################




class AutoGroupWord(AbstractGroupWord):
    """TODO: Yet to be documented     

    """

    def __init__(self,  *generators, **kwds):
        """Create word a in group from the generators given as arguments.

        Users should not refer to this class directly.
        """
        super(AutoGroupWord, self).__init__(*generators, **kwds)
        self.seq =  []
        self.reduced = 0
        for generator in generators:
            g = self.group.reduce_generator(generator)
            if g:
                self.seq.append(g)

    def __getitem__(self, i):
        return AutoGroupWord(self.group, self.seq[i])

    def __len__(self):
        return len(self.seq)

    def iter_generators(self):
        for generator in self.seq:
            yield generator

    def append_word(self, word):
        self.seq.extend(word.seq)
        

    def iter_inv(self):
        for generator in reversed(self.seq):
            yield from self.group.iter_inv(generator) 

    def str(self):
        return self.group.str_word(self)

    __repr__ = str

    def to_list(self):
        """Deprecated!

        For compatiblity with older functions only!
        """
        raise NotImplementedError("Deprecated")
        return [x.to_tuple() for x in self.seq]


####################################################################
### Class AutoGroup
####################################################################


ABORT_MANY_REWRITES = 50000
ABORT_MANY_REWRITES_TXT = (
  "Reduction in class AutoGroup aborted after %s rewrites" % (
                         ABORT_MANY_REWRITES))



class AutoGroup(AbstractGroup):   

    FRAME = re.compile(r"^(\w*)$")
    STR_FORMAT = "%s"

 
    word_type = AutoGroupWord  # type of an element (=word) in the group 

    def __init__(self, creator, rules, inverter):
        """ TODO: Yet to be documented     


        """
        self.rewrite_system = TaggedRewriteSystem(rules)
        self.substituter = self.rewrite_system.subst_word
        self.creator = creator
        self.inverter = inverter
        super(AutoGroup, self).__init__()


    ################################################################
    # There is no need to overwrite any methods below this line
    ################################################################


    def generator(self, *data):
        """Convert tuple *data to a group word"""
        if len(data):
            creator = self.creator[data[0]](*data)
            return self.word_type(*creator, group=self)
        return self.word_type(group=self)

        

    def reduce_generator(self, generator):
        return generator.reduce(self) if generator else None



    def str_word(self, g1):
        if len(g1.seq):
            return self.STR_FORMAT  % "*".join(map(str, g1.seq))
        return self.STR_FORMAT % "1"


    def reduce(self, word, copy = False):
        if word.reduced >= len(word.seq):
            return word
        if copy:
            word = self.copy_word(word) 
        w = word.seq
        gap = 0
        pos = word.reduced + 1
        min_gap = 64
        n_iterations = 0
        #print("start", w)
        while True:
            length, subst = self.substituter(self, w[:pos])
            #print("subst", length,  w[:pos], pos, subst, gap,  w[pos+gap:])
            if length:
                pos -= length
                gap += length - len(subst)
                if  gap < 0:
                    insert = max(min_gap, -gap)
                    w[pos:pos] = [None] * insert
                    gap += insert
                    min_gap *= 2
                subst = [self.reduce_generator(x) for x in subst]
                w[pos + gap : pos + gap + len(subst)] = subst 
            while pos + gap < len(w) and not w[pos + gap]:
                gap += 1
            if pos + gap >= len(w):
                w[pos:] = []
                word.seq = w
                word.reduced = pos
                #print("result", w)
                return word
            w[pos] = w[pos + gap]
            pos += 1
            #print("after:", w[:pos], pos, gap,  w[pos+gap:])
            n_iterations += 1 
            if n_iterations == ABORT_MANY_REWRITES:    
                warnings.warn(ABORT_MANY_REWRITES_TXT)
                w[pos:pos+gap] = []
                word.seq = w
                word.reduced = pos
                return word

    def copy_word(self, g1):
        """Return deep copy of group element g1"""
        result = self.neutral()
        result.seq = deepcopy(g1.seq)
        result.reduced = g1.reduced
        return result


    def _equal_words(self, g1, g2):
        """Return True iff elements g1 and g2 are equal 

        This method is called for elements g1 and g2 of the group
        'self' only.
        """
        g1 = self.reduce(g1, True)
        g2 = self.reduce(g2, True)
        tup1 = self.iter_to_tuples(g1)
        tup2 = self.iter_to_tuples(g2)
        equ = (x == y for x, y in zip(tup1, tup2))
        return len(g1) == len(g2) and all(equ)

        #tup1 = (x.to_tuple() for x in self.reduce(g1).iter_generators())
        #tup2 = (x.to_tuple() for x in self.reduce(g2).iter_generators())
        #equ = (x == y for x, y in zip(tup1, tup2))
        #return len(g1) == len(g2) and all(equ)


    def _imul(self, g1, g2):
        """Return product g1*g2 of group elements g1 and g2.

        g1 may be destroyed but not g2.

        This method is called for elements g1 and g2 of the group
        'self' only.
        """
        assert g1.group == g2.group == self, (g1.group, g2.group)
        g1.seq += g2.seq[:]
        return self.reduce(g1)


    def iter_inv(self, generator):
        """yield inverse of the generator

        The function must yield one or more generators of type
        TaggedAtom, so that the product of these generators is the 
        inverse of the input 'generator'.
        """
        yield from self.inverter[generator.tag](self, generator)
		
    def _invert(self, g1):
        """Return product g1**-1 for group elements g1.

        g1 must not be destroyed.

        This method is called for elements g1 of the group
        'self' only.
        """
        result = self.neutral()
        assert g1.group == self
        result.seq = [
           inv  for g1_generator in reversed(g1.seq)
               for inv in self.iter_inv(g1_generator)
        ]
        return result.reduce()


    def iter_to_tuples(self, g1):
        assert g1.group == self
        return (t for generator in g1.seq for t in generator.as_tuples())


    def as_tuples(self, g1):
        return list(self.iter_to_tuples(g1))




       



# cython: language_level=3

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from libc.stdint cimport uint32_t, uint16_t, uint8_t, int32_t
from libc.stdint cimport uint64_t
from libc.stdint cimport uint64_t as uint_mmv_t




######################################################################
### Wrappers for C functions from file mm_op255.pxd
######################################################################


cimport cython
cimport mm_op255

cimport cython

@cython.wraparound(False)
@cython.boundscheck(False)
def op_pi(v_in, delta, pi, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t delta_v_ = delta
    cdef uint32_t pi_v_ = pi
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op255.mm_op255_pi(&v_in_v_[0], delta_v_, pi_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_delta(v_in, delta, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t delta_v_ = delta
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op255.mm_op255_delta(&v_in_v_[0], delta_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_pi_tag_ABC(v, delta, pi, mode):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t delta_v_ = delta
    cdef uint32_t pi_v_ = pi
    cdef uint32_t mode_v_ = mode
    with nogil:
        mm_op255.mm_op255_pi_tag_ABC(&v_v_[0], delta_v_, pi_v_, mode_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_delta_tag_ABC(v, d, mode):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t d_v_ = d
    cdef uint32_t mode_v_ = mode
    with nogil:
        mm_op255.mm_op255_delta_tag_ABC(&v_v_[0], d_v_, mode_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_copy(mv1, mv2):
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    cdef uint_mmv_t[::1] mv2_v_ = mv2
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op255.mm_op255_copy(&mv1_v_[0], &mv2_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_compare_len(mv1, mv2, len):
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    cdef uint_mmv_t[::1] mv2_v_ = mv2
    cdef uint32_t len_v_ = len
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op255.mm_op255_compare_len(&mv1_v_[0], &mv2_v_[0], len_v_)
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_compare(mv1, mv2):
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    cdef uint_mmv_t[::1] mv2_v_ = mv2
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op255.mm_op255_compare(&mv1_v_[0], &mv2_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_vector_add(mv1, mv2):
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    cdef uint_mmv_t[::1] mv2_v_ = mv2
    with nogil:
        mm_op255.mm_op255_vector_add(&mv1_v_[0], &mv2_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_scalar_mul(factor, mv1):
    cdef int32_t factor_v_ = factor
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    with nogil:
        mm_op255.mm_op255_scalar_mul(factor_v_, &mv1_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_compare_mod_q(mv1, mv2, q):
    cdef uint_mmv_t[::1] mv1_v_ = mv1
    cdef uint_mmv_t[::1] mv2_v_ = mv2
    cdef uint32_t q_v_ = q
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op255.mm_op255_compare_mod_q(&mv1_v_[0], &mv2_v_[0], q_v_)
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_xy(v_in, f, e, eps, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t f_v_ = f
    cdef uint32_t e_v_ = e
    cdef uint32_t eps_v_ = eps
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op255.mm_op255_xy(&v_in_v_[0], f_v_, e_v_, eps_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_omega(v, d):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t d_v_ = d
    with nogil:
        mm_op255.mm_op255_omega(&v_v_[0], d_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_xy_tag_ABC(v, f, e, eps, mode):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t f_v_ = f
    cdef uint32_t e_v_ = e
    cdef uint32_t eps_v_ = eps
    cdef uint32_t mode_v_ = mode
    with nogil:
        mm_op255.mm_op255_xy_tag_ABC(&v_v_[0], f_v_, e_v_, eps_v_, mode_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_t(v_in, exp, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t exp_v_ = exp
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op255.mm_op255_t(&v_in_v_[0], exp_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_t_A(v_in, exp, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t exp_v_ = exp
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op255.mm_op255_t_A(&v_in_v_[0], exp_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_t_ABC(v, exp):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t exp_v_ = exp
    with nogil:
        mm_op255.mm_op255_t_ABC(&v_v_[0], exp_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_xi(v_in, exp, v_out):
    cdef uint_mmv_t[::1] v_in_v_ = v_in
    cdef uint32_t exp_v_ = exp
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    with nogil:
        mm_op255.mm_op255_xi(&v_in_v_[0], exp_v_, &v_out_v_[0])

@cython.wraparound(False)
@cython.boundscheck(False)
def op_xi_tag_A(v, exp):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t exp_v_ = exp
    with nogil:
        mm_op255.mm_op255_xi_tag_A(&v_v_[0], exp_v_)

@cython.wraparound(False)
@cython.boundscheck(False)
def op_word(v, g, len_g, e, work):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t[::1] g_v_ = g
    cdef int32_t len_g_v_ = len_g
    cdef int32_t e_v_ = e
    cdef uint_mmv_t[::1] work_v_ = work
    cdef uint32_t ret_
    with nogil:
        ret_ = mm_op255.mm_op255_word(&v_v_[0], &g_v_[0], len_g_v_, e_v_, &work_v_[0])
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_word_tag_A(v, g, len_g, e):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t[::1] g_v_ = g
    cdef int32_t len_g_v_ = len_g
    cdef int32_t e_v_ = e
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op255.mm_op255_word_tag_A(&v_v_[0], &g_v_[0], len_g_v_, e_v_)
    return ret_

@cython.wraparound(False)
@cython.boundscheck(False)
def op_word_ABC(v, g, len_g, v_out):
    cdef uint_mmv_t[::1] v_v_ = v
    cdef uint32_t[::1] g_v_ = g
    cdef int32_t len_g_v_ = len_g
    cdef uint_mmv_t[::1] v_out_v_ = v_out
    cdef int32_t ret_
    with nogil:
        ret_ = mm_op255.mm_op255_word_ABC(&v_v_[0], &g_v_[0], len_g_v_, &v_out_v_[0])
    return ret_


######################################################################
### Constants
######################################################################


MMV_ENTRIES = 247488

INT_BITS = 64

LOG_INT_BITS = 6

P = 255

FIELD_BITS = 8

LOG_FIELD_BITS = 3

INT_FIELDS = 8

LOG_INT_FIELDS = 3

P_BITS = 8

MMV_INTS = 30936





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import collections
import re
import warnings
from random import randint
import pytest


from mmgroup import mat24 as m24
from mmgroup.tests.spaces.sparse_mm_space import SparseMmV
from mmgroup.tests.groups.mgroup_n import MGroupNWord







p = 241
rep_mm = SparseMmV(p)
grp = MGroupNWord


def vector_data(v):
    data = v.as_tuples()
    assert len(data) == 1
    return data[0]


def as_suboctad(v1, d):
    c = m24.ploop_cap(v1, d)
    return m24.cocode_to_suboctad(c, d)


1234567890123456789012345678901234567890123456789012345678901234567890
def op_xy(v, eps, e, f):
    """Multiply unit vector v with group element

    This function multplies a (multiple of a) unit vector v
    with the group element

        g  =  d_<eps> * (x_<e>)**(-1)  * (y_<f>)**(-1) .
   
    It returns the vector w = v * g. This function uses the same 
    formula for calculating  w = v * g,  which is used in the 
    implementation of the monster group for computing
    v = w * g ** (-1).

    Input vector v  must be given as a tuple as described in class
    mmgroup.structures.abstract_mm_rep_space.AbstractMmRepSpace.
    Output vector is returned as a tuple of the same shape.
    """
    if len(v) == 3:
        v = (1,) + v
    v_value, v_tag, v_d, v_j = v
    parity = (eps >> 11) & 1  # parity of eps
    if v_tag == 'X':
        w_d = v_d ^ (f & 0x7ff)
        sign = v_d >> 12
        d = v_d & 0x7ff
        c =  m24.vect_to_cocode(1 << v_j)
        sign +=  m24.gcode_weight(e ^ f) 
        sign += m24.gcode_weight(f) 
        sign += m24.gcode_weight(d) * (parity + 1)
        sign += m24.gcode_weight(d ^ e ^ f)
        sign += m24.scalar_prod(e, c)
        cc = c if parity else 0 
        cc ^= eps ^ m24.ploop_theta(f) ^ m24.ploop_cap(e, f) 
        sign += m24.scalar_prod(d, cc)
        sign += f >> 12
        return (-1)**sign * v_value, 'X', w_d, v_j
    elif v_tag in 'YZ':
        tau = v_tag == 'Y'
        sigma = (tau + parity) & 1
        w_tag = "ZY"[sigma]
        sign = (v_d >> 12) + ((v_d >> 11) & tau)
        w_d = (v_d ^ e ^ (f & ~-sigma)) & 0x7ff
        #print("YZ, w_d", hex(w_d), hex(v_d ^ (e & 0x7ff) ^ (f & sigma_1)), err)
        # Next we check the sign
        d = v_d & 0x7ff
        c =  m24.vect_to_cocode(1 << v_j)
        sign += m24.ploop_cocycle(f, e) * (sigma + 1)
        sign += m24.gcode_weight(f) * (sigma)
        sign += m24.gcode_weight(d ^ e)  
        sign += m24.gcode_weight(d ^ e ^ f)  
        sign += m24.scalar_prod(f, c)
        cc = eps ^ m24.ploop_theta(e) 
        cc ^=  m24.ploop_theta(f) * ((sigma ^ 1) & 1)
        sign += m24.scalar_prod(d, cc)
        sign += (e >> 12) + (f >> 12) * (sigma + 1)
        sign += ((e >> 11) & sigma)
        return (-1)**sign * v_value, w_tag, w_d, v_j
    elif v_tag == 'T':
        d = m24.octad_to_gcode(v_d)
        w_j = v_j  ^ as_suboctad(f, d)
        sign = m24.gcode_weight(d ^ e) + m24.gcode_weight(e)
        sign +=  m24.scalar_prod(d, eps)
        sign += m24.suboctad_scalar_prod(as_suboctad(e ^ f, d), v_j)
        sign += m24.suboctad_weight(v_j) * parity
        return (-1)**sign * v_value, 'T', v_d, w_j
    elif v_tag in 'BC':
        m = v_tag == 'C'
        c = m24.vect_to_cocode((1 << v_d) ^ (1 << v_j))
        n = m ^ m24.scalar_prod(f, c)
        w_tag = "BC"[n]
        w_i, w_j = max(v_d, v_j), min(v_d, v_j)
        sign = m * parity + m24.scalar_prod(e ^ f, c)
        return (-1)**sign * v_value, w_tag, w_i, v_j
    elif v_tag == 'A':
        w_i, w_j = max(v_d, v_j), min(v_d, v_j)
        c = m24.vect_to_cocode((1 << v_d) ^ (1 << v_j))
        sign = m24.scalar_prod(f, c)
        return (-1)**sign * v_value,'A', w_i, v_j
    else:
        raise ValueError("Bad tag " + v_tag)




def one_test_op_xy(v, eps, e, f, verbose = 0):
    pi_atom = grp('d', eps)
    x_atom = grp('x', e)**(-1)
    y_atom = grp('y', f)**(-1)
    w = v * pi_atom * x_atom  * y_atom 
    if verbose:
        print("%s * %s * %s * %s = %s" % (v, pi_atom, x_atom, y_atom, w))
    v_tuple = vector_data(v)   
    w_tuple = vector_data(w) 
    w1_tuple = op_xy(v_tuple, eps, e, f)  
    w1 = v.space(v.p, [w1_tuple])
    if w != w1:
        print("%s * %s * %s * %s = %s" % (v, pi_atom, x_atom, y_atom, w))
        print("with xy formula:", w1)
        raise ValueError("Calculation with xy formula failed")







def rand_v(tag):
    return rep_mm(tag, "r")
   

def op_xy_testdata():
    data = [
       [ ("Y", 3, 6),  0, 0, 0x1171 ],
       [ ("Y", 3, 6),  12, 0, 0 ],
       [ ("Y", 3, 6),  12, 1111, 0 ],
       [ ("Y", 3, 6),  12, 0, 1111],
       [ ("Z", 3, 6),  0, 0, 0x1171 ],
       [ ("Z", 3, 6),  12, 0, 0 ],
       [ ("Z", 3, 6),  12, 1111, 0 ],
       [ ("Z", 3, 6),  12, 0, 1111],
       [ ("X", 3, 6),  0, 0, 0x1171 ],
       [ ("X", 3, 6),  12, 0, 0 ],
       [ ("X", 3, 6),  12, 1111, 0 ],
       [ ("X", 3, 6),  12, 0, 1111],
    ]
    data = []
    for v, f, e, eps in data:
        if isinstance(v, str): 
            v1 = rand_v(v)
        else :
            v1=  rep_mm(*v)
        yield v1,  f, e, eps
    for v in "ZY":
        for i in range(500):
            v1 =  rand_v(v)
            eps = randint(0, 0xfff) 
            e = randint(0, 0x1fff)
            yield v1,  eps, e, 0 
    v_tags = "ABCTYZX"
    #v_tags = "ZYX"
    for v in v_tags:
        for i in range(1000):
            v1 =  rand_v(v)
            eps = randint(0, 0xfff)
            e = randint(0, 0x1fff)
            f = randint(0, 0x1fff)
            yield v1,  eps, e, f
      
       
@pytest.mark.space
def test_op_xy(verbose = 0):
    print("Testing group operations x, y ...")
    for v, eps, e, f in op_xy_testdata():
        one_test_op_xy(v, eps, e, f, verbose = verbose)
    print("passed")









"""Parser for obtaining atoms of algebraic structures

This module contains classes an functions that support parsing of 
atoms of an algebraic structure from a string. In this module we
abuse the word "algebra" for denoting any algebraic structure.

Here an algebra is represented as subclass of an abstract class 
modelling an abstract algebra. Examples of such abstract classes 
are AbstractGroup modelling a group, and AbstractRepSpace modelling
a free finite-dimensional module over the integers modulo p with
a group operating on that module.

In order to make an abstract class AbstractXxx useful there are 
subclasses of class AbstractXxx, such that an instance of such a
subclass models a specific algebraic structure. Such a subclass
contains functions for performing the operations on the algebraic
structure. For each class AbstractXxx there is a class 
AbstactXxxWord modelling an element of a specific algebra. For 
an instance x of AbstactXxxWord the function x.algebra() returns 
the specific algebra to which x belongs. The main purpose of the 
class AbstactXxxWord is to overload standard python operators with 
the appropriate functions defined in the corresponding algebra. 
Since each class modelling an algebra contains functions to 
generate elements of that algebra, users need not deal with class 
AbstactXxxWord.


Many algebras are built up from atoms, e.g. groups may be built
up from generators and vector spaces may be built up from
multiples of unit vectors. There is a standard method to map
some strings to atoms for a specific algebra.

Here all atoms are tuples of the form

    (tag, data1, data2, ...)

where 'tag' is an alphabetic string of length 1 and 
data1, data2, ... are either nonnegative integers or alphanumeric
strings representing identifiers and not containing the undersore
character '_'.

We support a parser that reckognizes as string of shape

    tag_data1_data2[_ ...]

as an atom. Such a parser can distinguish between atoms, idenifiers
and numbers, where numbers are nonnegative integers.

The grammar for distingushing between atoms, identifiers and numbers 
in Extended Backus-Naur form is as follows:


literal           =  atom | identifer | number ;
atom              =  tag, "_" , index, {"_",  index} ;
identifier        =  (letter, (letter | digit), {alphanum}) ; 
number            =  decimal number | hex number ; 
tag               =  letter ;
index             =  num index | alpha index ;
num index         =  decimal number | hex index ;
hex index         =  hex number | (digit, {hex digit}, ("h" | "H")) ;
decimal number    =  digit, {digit} ;
hex number        =  "0x", hex digit, {hex digit} ;
alpha index       =  letter, {letter | digit} ;
alphanum          =  letter | digit | "_" ;
hex digit         =  digit | "a" ... "f" | "A" ... "F" ;
letter            =  "A" ... "Z" | "a" ... "z" ;
digit             =  "0" ... "9" ;

It turns out that the standard python parser reckognizes both,
atoms and identifiers as python identifers. 

Given a specific algebra, we may evaluate any python expression 
(using e.g. the python eval function) in such a way that all
python identifiers representing atoms are evaluated to the 
corresponding atoms of the specific algebra. Therefore we must
pass a mapping from the identifier strings to their values to
the evaluation function. Then the evaluation function parses
a string as a python expression, it maps all identifiers to
their value an returns the evaluated expression.

For an abstract algebra class AbstractXxx there may be an abstract 
class AtomicXxx modelling an algebra of type AbstractXxx generated
by atoms. We call such an algebra an 'atomic' algebra. As in the
case of abstract algebras, a specific algebra in modelled as an
instance of a subclass of class AbstractXxx. 

Calling a specific algebra with a single argument converts this
argument to an element of the algebra, as usual in python. If
the specific algebra is an atomic algebra, that parameter may be 
a python expression coded as a string. The idenifiers in that 
string, which have the shape of an atom according to the grammar
given above, are interpreted as atoms. This provides a simple
mechanism for generating atoms of an algebra.
 
"""

import re
import ast


#######################################################################
# Class AtomDict
#######################################################################


class AtomDict(object):
    """Models a parser for obtaining atoms as described above

    The constructor of this class takes two arguments :
    - a function 'reducer', 
    - a dictionary 'identifiers'.

    For an instance a of this class and a string s, a[s]
    parses the string a according to the grammer given above and
    returns one of the following objects:

    - an integer
    - the result of a function f(tag, data1, data2,...)
    - an object referred by a valid python identifier.

    If s is an integer recognized by the grammer given above
    then a[s] returns that integer.

    If s is a string of shape  tag_data1_data2[_ ...] as described
    by the grammar given above, then a[s] returns 
    reducer(tag, data1, data2, ...). Here 'tag' is a letter. i.e. a
    string of length 1, and the substrings data1, data2, ...  are
    evaluated to strings or to integers  according to the grammar
    given above.

    If s is an identifier contained as a key in the dictionary
    'identifiers' then a[s] returns identifiers[s].

    In any other case the invocation of a[s] raises ValueError.
    """
    match_atom = re.compile(r"[A-Za-z]_\w+", re.ASCII)
    match_identifier = re.compile(r"[A-Za-z][A-Za-z0-9]\w+", re.ASCII)
    match_number = re.compile(r"[0-9]+|0x[0-9A-Fa-f]+")
    match_letter = re.compile(r"[A-Za-z]")

    m_int = re.compile(r"([0-9]+)")
    m_hex1 = re.compile(r"([0-9][0-9A-Fa-f]*)[Hh]")
    m_hex2 = re.compile(r"(0x[0-9A-Fa-f]+)")
    m_random = re.compile(r"([A-Za-z][A-Za-z0-9]*)")

    m_dict_index = {
        m_int: int,
        m_hex1: lambda x : int(x,16),
        m_hex2: eval,
        m_random: str,
    }

    def __init__(self, reducer = tuple, identifiers = {}):
        self.identifiers = identifiers
        self.reducer = reducer

    def __getitem__(self, s):
        if self.match_atom.fullmatch(s):
            indices = map(self._parse_index, s[2:].split("_"))
            word = self.reducer(s[0], *indices)
            return word
        elif self.match_identifier.fullmatch(s):
            try:
                return self.identifiers[s]
            except KeyError:
                raise KeyError("Identifer %s not found in atom literal" %s)
        elif self.match_number.fullmatch(s):
            return eval(s)
        raise ValueError("Syntax error in atom literal")

    def _parse_index(self, s):
        for matobj, translator in self.m_dict_index.items():
            try:
                return translator(matobj.fullmatch(s).groups()[0])
            except:
                pass
        return ValueError("Syntax error in index of atom literal")

       


#######################################################################
# Evaluating a string with function eval_atom_expression()
#######################################################################


# For background on module ast (abstact syntax tree) see e.g.:
# https://greentreesnakes.readthedocs.io/en/latest/tofrom.html







# Set of nodes in module .ast considered as safe for function eval()
# See  https://greentreesnakes.readthedocs.io/en/latest/nodes.html
good_ast_nodes = set([
   ast.Expression,
   ast.Num, ast.Str, ast.Expr,
   ast.Load,
   ast.Constant,

   ast.UnaryOp,
   ast.UAdd, ast.USub, ast.Not, ast.Invert,

   ast.BinOp,
   ast.Add, ast.Sub, ast.Mult, ast.Div, ast.FloorDiv, ast.Mod, 
   ast.Pow, ast.LShift, ast.RShift, ast.BitOr, ast.BitXor, 
   ast.BitAnd,

   ast.Call,
   ast.Subscript, ast.Index 
])



try:
    good_ast_nodes.add(ast.NameConstant) # New in version 3.4
except:
    pass



class EvalNodeVisitor(ast.NodeVisitor):
    """Node visitor for Abstract Syntax Trees creeated by module  .ast 

    This vistor checks for all nodes of an AST whether they are secure 
    for processing with function eval(). Method visit() raises TypeError
    if an insecure node id found in the AST.

    Caution: 
    Leading underscores in python identifiers or attributes are illegal! 

    Note that leading underscores in an identifier or an attribute are
    considered dangerous. E.g.  parsing '().__class__.__bases__[0]'
    yields the class <object>, which a user should not manipulate. For  
    a discussion about the security of eval() see:
    https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html

    Creating tuples in backets or dictionaries in curly brackets is also
    illegal because the syntax gets messy when used in the code generator.
    """
    def visit(self, Node):
        cls = Node.__class__
        #print("visiting", Node, cls)
        if cls in good_ast_nodes:
            pass
        elif cls == ast.Name:
            if Node.id[:1] == "_":
                raise NameError("Illegal leading '_' in name in python expression") 
        elif cls == ast.Attribute:
            if Node.attr[:1] == "_":
                raise NameError("Illegal leading '_' in name in python expression") 
        else:
            print("\nIllegal ast node", type(Node))
            raise TypeError("Illegal construct in python expression") 
        self.generic_visit(Node)


def eval_atom_expression(s, atom_dict):
    """An attempt to create a secure version of function eval()

    See function EvalNodeVisitor() for allowd python contructs.
    """
    try:
        a_tree = ast.parse(s.strip(), mode="eval")
        EvalNodeVisitor().visit(a_tree)
        code = compile(a_tree, filename="atom expression", mode="eval")
        result = eval(code, {'__builtins__': {}},  atom_dict )
        return result
    except:
        print("\nError in evaluating an atom expression!\n")
        raise
        

#######################################################################
# Class TaggedAtom
#######################################################################

def ihex(x, length = None):
    x = int(x)
    sign = "-"[:(x < 0)]
    s = "%xh" % abs(x) if length is None else "%0*xh" % (length, abs(x))
    return sign + s if s[:1].isdigit() else sign + "0" + s

class TaggedAtom(object):
    """Models an atom of a general group with generators and relations"""

    # format_tuple[i] is used for formatting the i-th component.
    # Recommended entries of the tuple are str, hex, and ihex
    format_tuple = (str,)
    operator_char = "*"
      
    def __init__(self, tag, *data):
        """Construct an atom from a tag and a seqence of od data. 

        'tag' should be a single alphabetical character. All
        data should be nonnegative integers or alphanumeric
        strings with the first character not a digit.

        Subclasses may overwerite this method.
        """
        self.data = data
        assert isinstance(tag, str) and len(tag) == 1
        self.tag = tag
        
    def as_tuples(self):
        """Return representation of an atom as a list of tuples

        Subclasses may overwerite this method. The function
        may be implemented as a generator
        """
        yield (self.tag,) + self.data

    def reduce(self, *args):
        """Reduce atom 'self' to a simpler form if possible"""
        return self
    	
    def str_tuple(self, tuple_):
        strings = []
        for i, x in enumerate(tuple_):
            try:
                f = self.format_tuple[i]
            except IndexError:
                f = str
            strings.append(f(x))
        return "_".join(strings)

    def str(self):
        tuples = []
        for tuple_ in self.as_tuples():
            tuples.append(self.str_tuple(tuple_))  
        return self.operator_char.join(tuples)   

    __repr__ = str
    



#######################################################################
# Test program
#######################################################################


if __name__ == "__main__":
    def atom_to_tuple(*args):
        yield args
    def atom_to_part1(*args):
        yield args[1]
    def reducer(*data):
       return data[0]
    atom_dict = {"t": atom_to_tuple, "u" : atom_to_part1}
    a = AtomDict(reducer, atom_dict, {"x17": 17.5})
    print( a["t_1_a"], a["x17"])
    print( eval_atom_expression("(3 + u_1 ) * 100 * x17", a))
        


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import numpy as np
from random import randint
from mmgroup import MM0, Xsp2_Co1, AutPL, MM

import pytest


from mmgroup.mm import mm_group_prepare_op_ABC
from mmgroup.generators import mm_group_n_clear
from mmgroup.generators import mm_group_n_mul_element
from mmgroup.generators import mm_group_n_mul_word_scan
from mmgroup.generators import mm_group_n_to_word
from mmgroup.generators import mm_group_n_right_coset_N_x0

from mmgroup.clifford12 import xsp2co1_set_elem_word_scan
from mmgroup.clifford12 import xsp2co1_elem_subtype
from mmgroup.clifford12 import xsp2co1_elem_to_N0
from mmgroup.clifford12 import xsp2co1_mul_elem_word
from mmgroup.clifford12 import xsp2co1_elem_to_word


def display_elem(elem):
    g = Xsp2_Co1()
    np.copyto(g._data, elem)
    print(g)
    return g
    


def py_prepare_op_ABC(g):
    a = np.zeros(11, dtype = np.uint32)

    g = g.mmdata
    assert g.dtype == np.uint32
    has_t = has_l = reduce = False
    for x in g:
        reduce |= has_t
        tag, data = (x >> 28) & 7, x & 0xfffffff
        if tag == 5 and data: 
            has_t = True
        if tag == 6 and data: 
            has_l = True
        if tag == 7 and has_data:
             raise ValueError("Illegal tag")

    if not has_l:
        mm_group_n_clear(a)
        len_a = mm_group_n_mul_word_scan(a, g, len(g))
        assert len_a == len(g) 
        len_a = mm_group_n_to_word(a, a)
        return 1, a[:len_a]

    if reduce or len(g) > 11:
        gn = np.zeros(5, dtype = np.uint32)
        elem = np.zeros(26, dtype = np.uint64)
        pos = xsp2co1_set_elem_word_scan(elem, g, len(g), 0)
        assert 0 <= pos <= len(g)
        #display_elem(elem)
        mm_group_n_clear(gn)
        scan = mm_group_n_mul_word_scan(gn, g[pos:], len(g) - pos)
        assert scan >= 0
        assert pos + scan == len(g)
        
        if xsp2co1_elem_subtype(elem) == 0x48:
            res = xsp2co1_elem_to_N0(elem, a)
            assert res >= 0
            mm_group_n_mul_element(a, gn, a)
            len_a = mm_group_n_to_word(a, a)
            return 1, a[:len_a]
        else:
            e = mm_group_n_right_coset_N_x0(gn)
            len_a = mm_group_n_to_word(gn, gn)
            res = xsp2co1_mul_elem_word(elem, gn, len_a)
            assert res >= 0
            len_a = xsp2co1_elem_to_word(elem, a)
            assert 0 <= len_a <= 10
            if e:
                a[len_a] = 0x50000000 + e
                len_a += 1
            return 0, a[:len_a]

    return 0, np.copy(g)
            



def prep_op_ABC(g, must_be_in_N0 = False, verbose = 1):
    a = np.zeros(11, dtype = np.uint32)
    py_prepare_op_ABC(g)
    res = mm_group_prepare_op_ABC(g.mmdata, len(g.mmdata), a)
    if res < 0:
        print("Test function mm_group_prepare_op_ABC with g =")
        print(g)
        s = "Function mm_group_prepare_op_ABC returns with error %d" 
        raise ValueError(s % res)
    b0, len_a =  (res >> 8) & 1, res & 0xff
    a = a[:len_a]
    g1 = type(g)('a', a)
    return b0, a


def prep_op_ABC_testdata():
    for i in range(20):
        yield MM0('r', 'G_x0') * MM0('r', 'N_0'), False
    for i in range(20):
        yield MM0('r', 'N_0'), True

    pi0 = zip(list(range(10)), [x ^ 1 for x in range(10)])
    #print(pi0)
    g =  MM0([('l', 1), ('p', pi0), ('d', 0x800), ('l', 2), ('d', 0x800)])**5
    assert len(g.mmdata) > 11
    _ = Xsp2_Co1((MM0(g).reduce()))  # This should exist!!
    yield g, True
    


@pytest.mark.mm_op
def test_prepare_op_ABC(verbose = 0):
    print("Testing function mm_group_prepare_op_ABC")
    for i, (g, must_be_in_N0) in enumerate(prep_op_ABC_testdata()):
        b0, g1 = py_prepare_op_ABC(g)
        ok = g == g.group('a', g1)
        if must_be_in_N0:
            ok &= b0 == 1
        b02, g2 = prep_op_ABC(g, must_be_in_N0, verbose = verbose)
        #print(type(g2), type(g2), g1, g2)
        ok &= b02 == b0 and (g2 == g1).all()
        if  verbose or not ok:
            print("Input g:\n" + str(g))
            print("Py Output g (b0 = %d):\n" % b0 + str(g1))
            if must_be_in_N0 and not b0:
                print("b0 should be set but isn't")
            if (g2 != g1).any() or b02 != b0 :
                print("C Output g (b0 = %d):\n" % b02 + str(g2))
            if not ok:
                raise ValueError("Function py_prepare_op_ABC failed")
        b02, g2 = prep_op_ABC(g, must_be_in_N0, verbose = verbose)
       
    print("Test passed")








from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import collections
import re
import warnings
from numbers import Integral
from array import array
from random import randint



from mmgroup.generate_c import c_snippet, TableGenerator, make_table
from mmgroup.generate_c import UserDirective, UserFormat



from mmgroup import mat24

from mmgroup.dev.mm_basics.mm_tables import MM_OctadTable
from mmgroup.dev.mm_op.mm_op import MM_Op




###########################################################################
# Permutations of the 24 entries of a small array with a Benes network
###########################################################################


class Perm24_Benes(MM_Op):
    """Permute 24 small integers with a Benes network

    This is a modified version of class Perm24_Standard for permuting
    vectors of small integers.

    A modulus p for the small integers is given in the constructor of the 
    class.  An variable of type uint_mmv_t stores self.INT_FIELDS small
    integers, with each of them stored in a field of self.FIELD_BITS bits.
    Sizes self.INT_FIELDS and self.FIELD_BITS are calculated by a base
    class of this class depending on modulus p.

    Here a Benes network may be used in case self.INT_FIELDS >= 8 only,
    i.e. an integer of type uint_mmv_t muts be able to store at least
    8 small integers. The 24 small integers to be permuted will be 
    copied into a set of at most 3 local variables v0, v1, v2.

    The Benes network for performing the permutation can be computed by
    the C function mat24_perm_to_net(). The result of this function is
    a array B of type uint32_t[9] describing a Benes network for a 
    permutation. From array B we must compute another array M type
    uint_mmv_t[] and length at most 21, with the exact length given by
    self.n_benes_operations(). Array M is a more suitable mask for
    permuting fields of bitlength self.FIELD_BITS, see method
    self.iter_shift_op() for details.
 
    Method self.prepare(benes_net, result) generates code for computing
    array M from array B. Here benes_net is the name of a variable
    referring to the input array B and  result is the name of a variable
    referring to the output array M.

    The recommended way to use this class in the code generator is:

    // The following array contains the permutation to be performed
    uint8_t permutation[24];

    // The following code prepares the permutation operation
    uint32_t benes_net[9];     // Benes network
    // The following mask is used by the actual permutation code
    uint_mmv_t benes_mask[%{PERM24_BENES_MASK_LEN}]; 
    // Compute Benes network
    mat24_perm_to_net(permutation, benes_net);
    // Prepare mask array from Benes network
    // %%PERM24_BENES_PREPARE "benes_net", "benes_mask"

    // The following code permutes the small array 'source' and stores
    // the result to 'dest'. The result is negated if bit 0 of 'sign'
    // is set.
    uint_mmv_t *soure, *dest, sign;

    // Declare temporary variables t0, t1,...
    // %%PERM24_BENES_DECLARE "t"
    // Load 'source' to temporary variables
    // %%PERM24_BENES_LOAD "source"
    // Permute and possibly negate data in temp. variables
    // %%PERM24_BENES_PERMUTE "benes_mask", "sign"
    // Store temporary variables to 'dest'
    // %%PERM24_BENES_STORE "dest"

    Some details of the internal operation of the benes network are
    described in method iter_shift_cases.
    """

    BENES_SHIFTS = [1,2,4,8,16,8,4,2,1]
    permute_comment = """// Permute the 24 small integers in '{vars}' 
// using the Benes network given by '{mask}'. All small   
// integers are negated if bit 0 of 'sign' is set.
"""

    def __init__(self, p, prefer_benes_net = True):
        """Initialise for calulations with small integers modulo p

        p+1 must be a power of two. Calculations modulo p are described 
        in more detail in the base classes of this class.

        If 'prefer_benes_net' then self.use_benes_net is set True
        whenever Benes networks may be used for a given p.
        Ohterwise self.use_benes_net is set True only if using 
        a Benes network is an obvious advantage compared to using
        the methods of class Perm24_Standard.
        """
        super(Perm24_Benes, self).__init__(p)
        self.prep_table = []
        self.use_benes_net = (
            self.V24_INTS_USED <=  2 + bool(prefer_benes_net) )
        if self.V24_INTS_USED <= 3:
            self.prep_table = self.make_prep_table()
            #print("Benes table:", list(map(hex, self.prep_table)))
        self.tables.update(self.make_tables())
        self.directives.update(self.make_directives())

    def not_supported(self, *args, **kwds):
        err = "Benes network not supported for p=%d on a %d-bit system"
        raise TypeError(err % (self.P, self.INT_BITS)) 

    def n_benes_operations(self):
        """Return No of computation steps required for Benes network"""
        return len(self.prep_table)

    def iter_shift_cases(self):
        """Yield pairs (layer, offset) to be used for shift opration

        Given a permutation  p1 of 24 entries, the C function 
        mat24_perm_to_net() computes a Benes network with 9 layers for 
        that permutation. Each layer is  represented as an integer which 
        encodes an array of 24 bits. If bit i is in layer l of the 
        Benes network is set, then entry i must be exchanged with entry
        i + self.BENES_SHIFTS[l] in layer i. The first layer is layer 0.

        In layers 3, 4 and 5 bits are set at positions i < 8 only. 

        In the context of a small permutation, the 24 entries of a
        small array are stored in self.V24_INTS_USED integers of type
        uint_mmv_t, with 1 <= self.V24_INTS_USED <= 3. The integer at 
        offset y contains entries 

            y * INT_FIELDS, ..., y * INT_FIELDS + INT_FIELDS - 1 .

        This function yields pairs (layer, offset) describing a step.
        Steps are given in the order in which the Benes network will be 
        processed. Here offset = y in a step means that the integer at 
        offset y must be processed together with the integer at offset 
        y + k  for k = floor(self.BENES_SHIFTS[layer] / INT_BITS).
        This means that up to INT_FIELDS entries of a layer are
        processed simultaneously in one step. 

        The order of the operations is chosen so that we obtain a 
        good locality.
        """
        if self.V24_INTS_USED == 1: 
            for i in range(9): yield i, 0
        elif self.V24_INTS_USED == 2: 
            for ofs in [0,1]:
                for i in range(4 - ofs): yield i, ofs
            yield 4, 0
            for ofs in [0,1]:
                for i in range(5 + ofs, 9): yield i, ofs
        elif self.V24_INTS_USED == 3: 
            for ofs in [0,1,2]:
                for i in range(3): yield i, ofs
            for i in range(3,6): yield i, 0
            for ofs in [0,1,2]:
                for i in range(6, 9): yield i, ofs
        else:
            raise ValueError("Illegal parameter for Benes network")



    def make_prep_table(self):
        """Return the table required by method prepare()

        This is a compressed version of the table returend by method
        iter_shift_cases() with each pair (layer, offset) stored in
        on byte. Here the lower four bits of that byte store the value 
        'layer' and the higher four bits of that byte store the value
        offset/16. In case self.V24_INTS_USED == 3 the higher four 
        bits store the value offset/8 instead.
        """ 
        d = self.V24_INTS_USED == 3
        return [  (offset << (self.LOG_INT_FIELDS + d)) + layer
              for layer, offset in self.iter_shift_cases()]  

    def prep_declare(self):
        """Delare local and static variables for self.prepare()"""
        s = """uint_mmv_t tmp; 
uint_fast8_t i;
"""
        if self.V24_INTS_USED > 1:
            s +=  self.snippet("""static uint8_t tbl[] = {
    // %%TABLE table_prepare_perm24, uint8_t
};
"""     , table_prepare_perm24 = self.prep_table)
        return s


    def prep_step(self, benes_net):
        """Generate code to compute entry i of output array of self.prepare(). 

        """
        if self.V24_INTS_USED == 1: 
            s = "tmp = ({net})[i];"
        elif self.V24_INTS_USED == 2: 
            s = "tmp = tbl[i]; tmp = ({net})[tmp & 15] >> (tmp & 0xf0);" 
        elif self.V24_INTS_USED == 3: 
            s = "tmp = tbl[i];\n"
            s += "tmp = ({net})[tmp & 15] >> ((tmp & 0xf0) >> 1);" 
        return s.format(net = benes_net)

    def prepare(self, benes_net, result):
        """Generates code for the precomputation needed for a permutation

        Given an array B describing a Benes network for a permutation 
        of 24 elements, this method genrates code for the necessary
        precomputation for performing such permutations. The result of
        the precomputation is an array M of type uint_mmv_t[] and length
        at most 21, with the exact length given by 
        self.n_benes_operations().

        The use of array M is described in method self.iter_shift_op(). 

        Parameter 'benes_net'  is the name of a variable referring to
        input array B and parameter 'result' is the name of a variable
        referring to the output array M.

        The function generates a C block in curly brackets. That block
        may contain a static array of up to 21 bytes. It uses the
        identifiers 'i', 'tmp' and 'tbl' inside the block.
        """
        s = """{
    %{declare}
    for(i = 0; i < %{nsteps}; ++i) {
        %{step}
        // %%MMV_UINT_SPREAD tmp, tmp
        %{result}[i] = tmp;
    }
}
"""
        return self.snippet(s, declare = self.prep_declare(), 
            nsteps = len(self.prep_table), 
            step = self.prep_step(benes_net),
            result = result)


    def iter_shift_op(self):
        """Yield instructions for the actual Benes network operation

        This function yields a sequence of triples

            (offset1, offset2, shift)

        For the permuation of an array v of type uint_mmv_t[] representing 
        the small integers in bit fields of length self.FIELD_BITS
        this means:

        Exchange bit  i  in v[offset1] bit  i + shift  in v[offset2]
        if bit i in the corresponding mask is set.

        Method self.prepare() generates code for computing the 
        corresponding array of masks for a Benes network prepresenting
        a permutation of 24 small integers.

        As usual in a Benes network, we have  shift == 0  if 
        offset1 != offset 2.
        """
        for (layer, offset) in self.iter_shift_cases():
            diff, sh = divmod(self.BENES_SHIFTS[layer], self.INT_FIELDS)
            yield offset, offset+diff, sh*self.FIELD_BITS

    def cond_negate(self, var, sign):
        """Generate code to negate  array of small integers

        Parameter 'var' describes the set of  of 24 small integers to be
        negated as in method benes_permutation(). All these integers
        are negated if bit 0 of variable 'sign' is set. 'sign' must be 
        of type uint_mmv_t. 
        """
        s = "%{sign} = (0-(%{sign} & %{hex:1})) & %{smask:P, range(24)};\n"
        for i in range(self.V24_INTS_USED):
            if i == 1 and self.V24_INTS_USED == 2:
                s += "%{sign} &= %{smask:P, range(8)};\n"
            s += "%%{var}%d ^= %%{sign};\n" %  i
        return self.snippet(s, var = var, sign = sign) 


    def permute(self, mask, sign, var = None):
        """Generate code for permutation with Benes network.

        The generated code permutes a sequence of 24 small integers. All
        these integers are negated if bit 0 of the variable of type
        uint_mmv_t with name 'sign' is set. Variable 'sign' is destroyed.

        Parameter 'mask' is an array M of type uint_mmv_t[] describing the 
        permutation. 'var' describes a set of at most three variables of
        type uint_mmv_t containing the sequence of 24 small integers to  
        be permuted. If var="v", these variables have names v0, v1, v2.
        There are precisely self.V24_INTS_USED such variables. By default, 
        'var' is the last recent argument passed to method declare()

        The description 'mask' of the permutation should be generated 
        using the code generated by method prepare().
        """
        if var is None: var = self.last_var
        vars = ", ".join(var + str(i) for i in range(self.V24_INTS_USED))
        vars = "(" + vars + ")"
        s = self.permute_comment.format(vars=vars, mask=mask, sign=sign)
        s += self.cond_negate(var, sign)
        for i, (ofs1, ofs2, sh) in enumerate(self.iter_shift_op()):
            if sh:
                assert ofs1 == ofs2
                s +=  """{t} = ({v}{ofs1} ^ ({v}{ofs1} >> {sh})) & {m}[{i}]; 
{v}{ofs1} ^=  {t} ^ ({t} << {sh});
""".format( t = sign, v = var, ofs1 = ofs1, sh = sh, m = mask, i = i )
            else: 
                s += """{t} = ({v}{ofs1} ^ {v}{ofs2}) & {m}[{i}]; 
{v}{ofs1} ^=  {t};  {v}{ofs2} ^=  {t};
""".format( t = sign, v = var, ofs1 = ofs1, ofs2 = ofs2, m = mask, i = i)
        s += "// Permutation of small integers done.\n"
        return s

    def declare(self, var):
        vars = ", ".join(var + str(i) for i in range(self.V24_INTS_USED)) 
        self.last_var = var
        return "uint_mmv_t %s;\n" % vars

    def load(self, source, var = None):
        if var is None: var = self.last_var
        s = ""
        for i in range(self.V24_INTS_USED):
            s += "{var}{i} = ({source})[{i}];\n".format(
                 source = source, var = var, i = i)
        return s;

    def copy(self, var_src, var_dest):
        s = ""
        for i in range(self.V24_INTS_USED):
            s += "({dest})[{i}] = ({src})[{i}];\n".format(
                 src = var_src, dest = var_dest, i = i)
        for i in range(self.V24_INTS_USED, self.V24_INTS):
            s +=  "({dest})[{i}] = 0;\n".format(
                    dest = var_dest, i = i)
        return s;

    def store(self, destination, var = None):
        if var is None: var = self.last_var
        s = ""
        for i in range(self.V24_INTS_USED):
            s += "({destination})[{i}] = {var}{i} ;\n".format(
                 destination = destination, var = var, i = i)
        for i in range(self.V24_INTS_USED, self.V24_INTS):
            s +=  "({destination})[{i}] = 0;\n".format(
                    destination = destination, i = i)
        return s;

    def make_tables(self):
        return {
            "PERM24_USE_BENES_NET": self.use_benes_net,
            "PERM24_BENES_MASK_LEN": len(self.prep_table),
        }
 
    def make_directives(self):
        d = {
           "PERM24_BENES_PREPARE" : UserDirective(self.prepare, "ss"),
           "PERM24_BENES_DECLARE" : UserDirective(self.declare, "s"),
           "PERM24_BENES_LOAD" : UserDirective(self.load, "ss"),
           "PERM24_BENES_COPY" : UserDirective(self.copy, "ss"),
           "PERM24_BENES_PERMUTE" : UserDirective(self.permute, "sss"),
           "PERM24_BENES_STORE" : UserDirective(self.store, "ss"),
        }
        if len(self.prep_table):
            return d
        else:
            d0 = {}
            for key_ in d: d0[key_] = self.not_supported
            return d0
        


###########################################################################
# Scalar products and inversions of a 2048 x 24 vector
###########################################################################


class ScalarProd2048(MM_Op):

    def __init__(self, p):
        super(ScalarProd2048, self).__init__(p)
        self.tables.update(self.make_tables())
        self.directives.update(self.make_directives())

    ######################################################################
    # Generate tables for scalar products
    ######################################################################

    def scalar_prod_table_entry(self, v1):
        v = mat24.gcode_to_vect(v1)
        return [ self.smask(self.P, v >> j)
                     for   j in range(0, 32, self.INT_FIELDS) ]

    def scalar_prod_table(self, dist, length, start=0):
        return sum( (self.scalar_prod_table_entry(v1)
           for v1 in range(start, start + length*dist, dist)), [] )

    def scalar_prod_2048_unroll(self, p_high, p_low, v):
        s = "uint_mmv_t v_t;\n"
        e1_3 = [self.scalar_prod_table_entry(i) for i in range(1,4)]
        for i in range(self.V24_INTS_USED):
            s += "{v}[{i}] ^= (v_t = {ph}[{i}] ^ {pl}[{i}]);\n".format( 
                 v = v, ph = p_high, pl = p_low, i = i)
            for (j, e) in tuple(zip(range(1,4), e1_3)):
                s += "{v}[{k}] ^= v_t ^ {e};\n".format( v = v,
                   k = j*self.V24_INTS + i, e = self.hex(e[i]) )
        return s



    def make_directives(self):
        return {
            "SCALAR_PROD_2048_UNROLL": 
                   UserDirective(self.scalar_prod_2048_unroll, "sss")
        }

    def make_tables(self):
        return {
            "MMV_TBL_SCALPROD_HIGH": self.scalar_prod_table(64, 32),
            "MMV_TBL_SCALPROD_LOW": self.scalar_prod_table(4, 16),
        }



###########################################################################
# Permutations of 64 entries in a row
###########################################################################


def v2_list(log_n):
    v2list = []
    for i in range(log_n):
        v2list = v2list + [i] + v2list
    return [None] + v2list
     

class SmallPerm64(MM_Op):
    N_REGS = 3
    v2 = v2_list(6)

    def __init__(self, p, shift = True):
        super(SmallPerm64, self).__init__(p)
        self.directives.update(self.make_directives())



    def generate_load_perm64(self, source, bytes, sign, type_ = ""):
        """Load 64 enries of a vector 'source' to a byte array 'bytes'

        The vector of type uint_mmv_t[] is negated if bit 12 of the
        'sign' is set.
        """
        s = """
// Load 64 entries of vector %s to byte array %s.
// The byte array is negated if bit 12 of %s is set.
""" % (source, bytes, sign)
        n_registers = 8 // self.FIELD_BITS
        registers =  ", ".join(["r%d" % i for i in range(n_registers)])
        s = "uint_mmv_t %s;\n" % registers
        LOAD_MASK = self.hex(self.smask(self.P, -1, 8))
        SIGN_MASK = self.hex(self.smask(self.P, -1))
        s += "%s = (0-((%s >> 12) & 1)) & %s;\n" % (sign, sign, SIGN_MASK)
        type_ = "(%s)" % type_ if len(type_) else ""
        for i, start in enumerate(range(0, 64, self.INT_FIELDS)):
            s += "r0 =  (%s)[%d] ^ (%s);\n" % (source, i, sign)
            for j in range(1, n_registers):
                s += "r%d = (r0 >> %d) & %s;\n" % (
                   j, j * self.FIELD_BITS, LOAD_MASK)
            s += "r0 &= %s;\n" % LOAD_MASK
            for j in range(self.INT_FIELDS):
                s += "(%s)[%d] = %s(r%d >> %d);\n" % (bytes,  start + j, 
                    type_, j % n_registers, 8 * (j // n_registers))
        return s;

    def setup_store(self, op):
        N_REGS = max(self.N_REGS, self.LOG_INT_FIELDS)
        indices =  ["r%d" % i for i in range(N_REGS)]   
        registers = ["ri"] +  indices          
        indices += ["%s[%d]" % (op, i) for i in range(N_REGS,6)]
        s = "uint_mmv_t v;\n"
        s +=  "uint_fast8_t %s;\n" % ", ".join(registers);
        for i in range(N_REGS):
            s += "r%d = %s[%d];\n" % (i, op, i)
        s += "ri = %s;\n"  % indices[0]
        return s, indices


    def process_field(self, bytes, i, dest):
        s = "(uint_mmv_t)(%s[%s])" % (bytes, "ri" if i else "0")
        i0 = i &  (self.INT_FIELDS - 1)
        s = "v %s= %s << %d;\n" % ("+" if i0 else "",  
               s,  i0 << self.LOG_FIELD_BITS)
        if i0 == self.INT_FIELDS - 1:
            s += "%s[%d] = v ;\n" % (dest, i >> self.LOG_INT_FIELDS)
        return s     
        
    def modify_index(self, i, indices):
        if i < 2: return ""
        return "ri ^= %s;\n" % indices[self.v2[i]]

    def generate_store_perm64(self, bytes, dest, op):
        s, indices = self.setup_store(op)
        for i in range(64):
            s += self.modify_index(i, indices)
            s += self.process_field(bytes, i, dest)
        return s

    def generate_invert_perm64(self, src):
        s = ""
        wt = MM_OctadTable.perm64_weights
        for i in range(self.V64_INTS):
            s += self.snippet("%{src}[%{i}] ^= %{smask:P,{mask}};\n",
                src = src, i = i,  mask = wt >> (i * self.INT_FIELDS))
        return s


    def _test(self):
        s = "#include <stdint.h>\ntypedef uint64_t uint_mmv_t;\n"
        s += "void f(int src[], int dest[], int op[])\n{\n"
        s += self.generate_load_perm64("src", "dest", "op") 
        s += "}\n"
        return s


    def make_directives(self):
        return {
           "LOAD_PERM64" : UserDirective(self.generate_load_perm64, "ssss"),
           "STORE_PERM64" : UserDirective(self.generate_store_perm64, "sss"),
           "INVERT_PERM64" : UserDirective(self.generate_invert_perm64, "s"),
        }



######################################################################
# Test functions
######################################################################





if __name__ == "__main__":
    c = Perm24_Standard(7)
    print(c.prepare("perm", "perm_fast"))
    for i, x in c.expressions("source", "perm_fast", "sign"):
          print("dest[%s] =\n  %s;" % (i, x))
    print("1234567890"*7)
    print("\n//\n")

    c = Perm24_Benes(7)
    print(c.prepare("net", "result"))
    print(c.declare("v"))
    print(c.load("source"))
    print(c.permute("maask", "sign"))
    print(c.store("dest"))
    1/0


    with open("tmp.c", "wt") as f: f.write(SmallPerm64(3)._test())
    os.system("gcc -c tmp.c")
    os.system("del tmp.c tmp.o")
    1/0

    for p in [3, 7, 127]:
        print ("\ntest", p)
        a = SmallPerm24(p)
        a._test()








"""Test C function dealing with Leech lattice vectors mod 3 of type 2

In this script we test functions dealing with vectors of type 2 in the 
Leech lattice modulo 3. These functions are implemented in file 
gen_leech3.c and available in the extension mmgroup.generators.

We use the terminology defined in
the document *The C interface of the mmgroup project*, 
section *Description of the mmgroup.generators extension*.
"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint, choices #, shuffle, sample
from functools import reduce
from operator import __or__
from numbers import Integral
from collections import defaultdict
from multiprocessing import Pool, TimeoutError

import numpy as np
import scipy
import scipy.special
from scipy.stats import chisquare
import pytest

from mmgroup import MM0
from mmgroup.mat24 import MAT24_ORDER, ploop_theta
from mmgroup.mat24 import bw24 as mat24_bw24
from mmgroup.generators import gen_leech3to2_short
from mmgroup.generators import gen_leech2to3_short
from mmgroup.generators import gen_leech3_neg
from mmgroup.generators import gen_leech2_type
from mmgroup.generators import gen_leech2_subtype
from mmgroup.generators import gen_leech2_type2

from mmgroup.mm import mm_aux_index_extern_to_sparse
from mmgroup.mm import mm_aux_index_sparse_to_leech2
 

from mmgroup.tests.test_gen_xi.test_gen_type4 import str_v3


#####################################################################
# Creating test vectors
#####################################################################




def type2_testvectors():
    """Yield short test vectors in the Leech lattice mod 2

    The function yields pairs ``(v2, vtype)``, where ``v2 ``is a short
    vector (of type 2) in the Leech lattice mod 2 in **Leech lattice
    encoding**, and ``vtype`` is the subtype of ``v``.

    This function uses function from file mm_aux.c for converting
    random indices of the rep 196884x to short vectors.
    """
    TEST_PARS = [
        (300, 2 * 276, 50, 0x20),
        (852, 759 * 64, 500, 0x22),
        (49428, 2048 * 24, 500, 0x21),
    ] 
    for (start, nvectors, ntests, vtype) in TEST_PARS:
         for i in range(ntests):
             v_extern = randint(start, start + nvectors - 1)
             v_sparse = mm_aux_index_extern_to_sparse(v_extern)
             v2 = mm_aux_index_sparse_to_leech2(v_sparse)
             assert v2 != 0, (v_extern, hex(v_sparse), hex(v2))
             #print(v_extern, hex(v_sparse), hex(v2), hex(vtype))
             yield v2, vtype


#####################################################################
# Test mapping of type 2 Leech vectors modulo 3 to vectors modulo 2
#####################################################################


def v3_to_v2(v3):
    """Map type-2 vector from Leech lattice mod 3 to Leech lattice mod 2

    Here parameter ``v3`` is a type-2 vector in the Leech lattice mod 2 
    in Leech lattice mod 3 encoding. 

    The function returns a type-2 vector in the Leech lattice mod 2
    corresponding to ``v3`` in Leech lattice encoding.

    The result is unique. The function returns 0 if ``v3`` is not of
    type 2 in the Leech lattice mod 3
    
    Tis function is a wrapper for the C function ``gen_leech3to2_short`` 
    in file ``gen_leech3.c``. 
    """
    result = gen_leech3to2_short(v3)
    assert result != 0, (str_v3(v3), weight_v3(v3), hex(result))
    return result

def v2_to_v3(v2):
    """Map type-2 vector from Leech lattice mod 2 to Leech lattice mod 3

    This function reverses the peration of function ``v3_to_v2``.
    The result is unique up to sign only.
    
    Tis function is a wrapper for the C function ``gen_leech2to3_short`` 
    in file ``gen_leech3.c``. 
    """
    result = gen_leech2to3_short(v2)
    assert result != 0, (hex(v2), hex(result))
    return result


@pytest.mark.gen_xi
def test_type3(verbose = 0):
    r"""Test conversion of type-2 vectors 

    The function converts random sshort vectors in the Leech
    lattice mod2 2 to vectors in the Leech lattice mod 3, and
    vice versa, and check that these conversions are consistent. 

    It also checks functions in module ``checks  
    """
    weights = defaultdict(int)
    for ntest, (v2, vtype) in enumerate(type2_testvectors()):
        if verbose:
            print("\nTEST %s" % (ntest+1))
            print("v2 = " , hex(v2))
        assert gen_leech2_subtype(v2) == vtype
        assert gen_leech2_type2(v2) == vtype

        v3 = v2_to_v3(v2) 
        assert v3_to_v2(v3) == v2
        v3n = gen_leech3_neg(v3)
        assert v3_to_v2(v3n) == v2
    
    
    



r"""

The python extension ``mmgroup.generators`` provides a random
generator. Its main purpose is the very fast generation of
a vector of integers modulo :math:`p` for the representation
:math:`\rho_p` of the monster group for small numbers :math:`p`.

The internal operation of the random generator is described in
file ``gen_random.c``. Here we describe the python interface
of the random generator. 

This interface is given by a set of functions that can be
imported from python module ``mmgroup.generators``.

A python function dealing with the randon generator requires
a seed. Here a seed is either an object returned by function
``rand_make_seed`` or ``None`` (default).

The default seed is (hopefully) thread save, and it is
initialized from volatile sources such as the time, the process
and the thread id, etc.

If hundreds of threads use the default seed, it may be useful
to call function ``rand_set_seed(None)`` from time to time.

A seed created by function ``rand_make_seed`` may be initalized
from a volatile source or from a fixed source (which is a 
64-bit integer). Such a seed may be used by one thread only.


Python functions in module ``mmgroup.generators``
.................................................


The following functions should be imported from
module ``mmgroup.generators``.


.. py:function:: .rand_get_seed(seed = None)
    :noindex:
 
    Return seed associated with parameter ``seed`` 

    This function is used for obtaining a suitable python
    object for calling a C function that deals with the 
    random generator. In case ``seed = None`` (default) it 
    returns the default seed object.

.. py:function:: .rand_make_seed(value = None)
    :noindex:

    Create a seed object for the random generator

    The function creates a seed object and returns that object.
    It is intialized with parameter ``value``, which must be
    ``None`` (default) or an unsigned 64-bit integer.

    By default, the seed is initilized from a volatile source. 

    The returned seed object is not thread safe.
 

.. py:function:: .rand_set_seed(seed = None, value = None)
    :noindex:

    Reseed an existing seed object

    The function reseeds an existing seed object ``seed``
    with a ``value``.

    Parameter ``seed`` may be ``None`` (default, referring to 
    the default seed) or an existing seed object created
    by function ``rand_make_seed``.

    Parameter ``value`` is a value for reseeding a seed object
    as in function ``rand_make_seed``. 

    In case ``seed = None`` (default) parameter ``value``
    must also be ``None`` (default).

    If many threads use the default seed then one default
    seed object is created per thread. There is no easy way to 
    determine when a thread has terminated. Here calling
    ``rand_set_seed(None)`` gives all these seed objects
    to the garbage collector, and new default seed objects 
    will be created on demand.


.. py:function:: .rand_bytes_modp(p, num_bytes, seed = None)
    :noindex:

    Return a random array of integers mod p
  
    The function returns a one-dimensional numpy array 
    of uniform distributed random integers ``x`` with 
    ``0 <= x < p``. That array has length ``num_bytes``
    and has ``dtype = numpy.uint8``. 

    Parameter ``seed`` must be either ``None`` (default, 
    referring to the default seed) or a seed object
    created by function ``rand_make_seed``.


.. py:function:: .rand_fill_bytes_modp(p, array_bytes, seed = None)
    :noindex:

    Fill an existing array with random integers mod p
  
    The function fills the one-dimensional numpy array ``array_bytes``
    with  uniform distributed random integers ``x`` with ``0 <= x < p``. 
    That array must have ``dtype = numpy.uint8``. 

    Parameter ``seed`` must be either ``None`` (default, 
    referring to the default seed) or a seed object
    created by function ``rand_make_seed``.


.. py:function:: .rand_gen_modp(p, seed = None)
    :noindex:

    Return random integer ``x`` with ``0 <= x < p``.

    Here 1 <= p <= 2**32 must hold.

    Parameter ``seed`` must be either ``None`` (default, 
    referring to the default seed) or a seed object
    created by function ``rand_make_seed``.


"""



r"""



The functions in file ``xsp2co1.c`` implement the group operation 
of the subgroup 
:math:`G_{x0}` (of structure :math:`2^{1+24}.\mbox{Co}_1`)
of the monster. 

Represenation of :math:`G_{x0}` on the tensor product :math:`4096_x \otimes \Lambda`
....................................................................................

In :cite:`Seysen20`, section 7.4  and 9, the generators  
:math:`x_d, x_\delta, y_\delta, x_\pi, \xi` of :math:`G_{x0}`
are also defined as generators of a group  
:math:`G_{x1} \subset  G(4096_x) \times  G(24_x)`. Here 
:math:`G(4096_x)` is a subgroup of the real Clifford group
:math:`\mathcal{C}_{12}` operating on the 4096-dimensional 
real vector space :math:`4096_x`, and :math:`G(24_x)` is
the automorphism group  :math:`\mbox{Co}_0` of the Leech 
lattice :math:`\Lambda`. :math:`G_{x1}` is a preimage of 
:math:`G_{x0}` with :math:`|G_{x1}:G_{x0}| = 2`, and 
:math:`G_{x0}` operates faithfully on the tensor product 
:math:`4096_x \otimes_\mathbb{Z} \Lambda`. Let 
:math:`(x_g, L_g) \in G_{x1} \subset G(4096_x) \times  G(24x)`.
Component :math:`x_g` determines component :math:`L_g` of
the pair up to sign. So it suffices to store the image 
:math:`v_g = v_0 \cdot L_g` of a fixed shortest vector 
:math:`v_0 \in \Lambda / 3 \Lambda` instead 
of the whole automorphism :math:`L_g` of :math:`\Lambda`.
We put :math:`v_0 = (0,0,4,-4,0,...,0)`  in the standard basis 
of the Leech lattice.

We can reconstruct :math:`L_g` from   :math:`x_g` and 
:math:`v_g` as follows. :math:`G(4096_x)` has an extraspecial 
subgroup :math:`Q_{x1}` of structure :math:`2^{1+24}`. The 
quotient of :math:`Q_{x1}` by its center :math:`Z(Q_{x1})` is 
isomorphic to :math:`\Lambda / 2 \Lambda`. By definition of 
:math:`G_{x1}`, the element :math:`L_g` operates on 
:math:`\Lambda / 2 \Lambda` in the same way as  :math:`x_g` 
operates on :math:`Q_{x1}/Z(Q_{x1})` by conjugation.
So that operation of :math:`L_g` can be reconstructed from
:math:`x_g`. A short vector in  :math:`\Lambda / 2 \Lambda`
has precisely two short preimages  in  :math:`\Lambda` which
are opposite. Thus the image of one short vector is known
exactly, and the images of all short vectors are known up to 
sign. Since :math:`L_g` preserves the scalar product, the 
images of all short vectors in :math:`\Lambda` can be computed. 
Function ``xsp2co1_elem_to_leech_op`` in file ``xsp2co1.c``
constructs :math:`L_g`  from  :math:`x_g` and :math:`v_g`.
The functions in file ``gen_leech.c`` support operations on
:math:`\Lambda / 2 \Lambda` and :math:`\Lambda / 3 \Lambda`. 


Embedding a group related to :math:`G_{x0}` into a Clifford group
.................................................................

The representation :math:`4096_x` of the subgroup  
:math:`G(4096_x)` of :math:`\mathcal{C}_{12}` extends to 
the standard representation of the Clifford group 
:math:`\mathcal{C}_{12}`. So we may use the theory in 
section :ref:`clifford-group-label` for computing in 
:math:`G(4096_x)`. In that section the basis vectors of the
standard representation  of :math:`\mathcal{C}_{12}` are 
numbered from :math:`0` to :math:`2^{12}-1`. We will 
assign compatible numbers to the basis vectors of the
representation :math:`4096_x` . 


Let  :math:`\mathcal{P}` be the Parker loop as defined in 
section :ref:`parker-loop-label`. The vector space 
:math:`4096_x` has basis vectors 
:math:`d_1^+, d_1^-, d \in \mathcal{P}`, with  relations 
:math:`(-d)_1^\pm = -d_1^\pm`, :math:`(\Omega  d)_1^+ = d_1^+`, 
:math:`(\Omega  d)_1^- = -d_1^-`. 
In the same section the elements of :math:`\mathcal{P}` are 
numbered from :math:`0` to :math:`2^{13}-1`. For 
:math:`0 \leq d < 2^{11}` we identify the 
basis vectors :math:`d_1^+` and :math:`d_1^-` of  :math:`4096_x`
with the basis vectors with numbers corresponding to :math:`d` 
and to :math:`d+2^{11}` of the representation of 
:math:`\mathcal{C}_{12}`, respectively. 



So we may represent an element :math:`x_g` of :math:`G(4096_x)`
as a **quadratic state matrix** in a structure of type 
``qstate12_type``, as defined in file ``clifford12.h``. Then
we may use the functions in file ``qmatrix12.c`` for computing
in :math:`G(4096_x)`.



The :math:`G_{x0}` representation 
.................................. 

We actually represent an element of :math:`G_{x0}` in **G_x0 
representation**. This is as an array ``elem`` of 26 integers 
of type ``uint64_t``. That array contains a pair
:math:`(x_g,v_g)` as described above with :math:`v_g` stored 
in the first entry ``elem[0]``, and  :math:`x_g` stored in the 
remaining 25 entries of array ``elem``.

Vector :math:`v_g \in \Lambda / 3 \Lambda` is stored in
``elem[0]`` as a 48-bit integer in **Leech lattice mod 3 
encoding** as described in the documentation of module 
``gen_leech.c``.

We store the inverse :math:`x_g^{-1}` of  :math:`x_g` in
the upper 25 entries of the array ``elem``. The rationale
for storing the inverse is that this simplifies the operation
of :math:`G(4096)_x` on :math:`Q_{x1}` by conjugation, using
function ``qstate12_pauli_conjugate`` in file ``qmatrix12.c``.
Note that :math:`x_g^{-1} = x_g^\top`, since  :math:`x_g`
is orthogonal.


In section :ref:`clifford-group-label` an element :math:`c`
of :math:`\mathcal{C}_{12}` is given as a real
:math:`2^{12} \times 2^{12}` matrix stored in a structure
of type ``qstate12_type``. In case :math:`c \in G(4096_x)`
that matrix has rational entries, where the denominators are 
powers of two. A structure of type ``qstate12_type`` 
representing a  :math:`c \in G(4096_x)` contains 
a triple :math:`(e,A,Q)`. There :math:`e` is a signed integral  
power of two, and :math:`(A,Q)` is a pair of bit 
matrices with up to 25 rows and up to 49 colums, where the 
first 24 columns belong to matrix :math:`A`, and the remaining 
columns belong to  matrix :math:`Q`.


Changing the sign of :math:`e` corresponds to negation of a 
matrix :math:`x_g^{-1} \in G(4096_x)` given by :math:`(e,A,Q)`
or to  multiplication by :math:`x_{-1}`. In the group 
:math:`G_{x0}`, changing the sign of :math:`v_g` has the same 
effect as changing the sign of :math:`x_g`; so  :math:`e` can 
always be made positive. The absolute value of :math:`e` is 
just there for scaling the operator norm of a matrix 
:math:`c` to 1, and can be omitted.

So we can represent one of the values :math:`\pm x_g^{-1}` as 
a pair :math:`(A,Q)`  with an implied positive scalar factor 
:math:`e` as above, and negate component :math:`v_g` of the 
pair :math:`(x_g,v_g)`, if necessary. 
 
Assuming that  :math:`x_g^{-1}` is stored in a structure 
``qs`` of type ``qstate_type``, we copy all valid entries 
``qs.data[i]`` to ``elem[i+1]``. This amounts to copying the
components  :math:`(A,Q)` of the triple  :math:`(e,A,Q)`
from ``qs`` to ``elem``.  We always reduce the quadratic
state matrix in ``qs`` before copying it, as indicated in 
section :ref:`clifford-group-label`. We fill unused entries 
and bits in the array ``elem`` with zeros. Thus  the 
representation of a :math:`g \in G_{x0}` in memory is unique.
``qs`` can easily be reconstructed from ``elem``, assuming
that the scalar factor :math:`e` in the triple
:math:`(e,A,Q)` is positive, and that trailing zero 
entries in the array ``elem`` are unused.

A word :math:`w` in the generators of the subgroup :math:`G_{x0}` 
of the monster group can be converted to **G_x0 representation**
by calling function ``xsp2co1_set_elem_word`` in file 
``xsp2co1.c``. Here the word :math:`w` must be given as an array `
of unsigned 32-bit integers as described in section 
:ref:`header-mmgroup-generators-label`.

The normal subgroup :math:`Q_{x0}` of :math:`G_{x0}`
.....................................................

The group  :math:`G_{x0}` also has an extraspecial normal
subgroup :math:`Q_{x0}` of structure :math:`2^{1+24}`. There is a 
natural isomorphism between the normal subgroup  :math:`Q_{x1}`
of :math:`G(4096_x)` and  :math:`Q_{x0}`. Considering
:math:`G(4096_x)` and  :math:`G(24_x)` as matrix groups,
this isomorphism is  given by:

.. math::
    x \in Q_{x1} \, \longmapsto \,  x \otimes 1 \in 
    G(4096_x) \otimes  G(24_x) \cong G_{x0} \, .

We identify :math:`Q_{x0}` with
:math:`Q_{x1}`. Both groups are generated by elements 
:math:`x_d, x_\delta`, as described above. 

We also represent an element of :math:`Q_{x0}` as a 25-bit
integer in **Leech lattice encoding** as described in 
section :ref:`mmgroup-generators-label`. 


Changing the basis of the space :math:`4096_x`
...............................................


We also use the vectors :math:`(d'), d \in \mathcal{P}` as basis 
vectors of :math:`4096_x` , where:

.. math::
    (d)' = \frac{1}{\sqrt2} \left( d_1^+ + d_1^- \right) \, , \quad 
    d_1^\pm = \frac{1}{\sqrt2} \left( (d)' \pm (\Omega d)' \right) \, . 

    
The reason for introducing this basis is that the operation of 
:math:`G_{x0}` on  :math:`Q_{x0}` (by conjugation) is
easy when we use this basis.
Then we obviously have :math:`(-d)' = -(d)'`. We can transform 
the coordinates of a vector :math:`v` from the basis  given by
:math:`d^\pm`  to the basis given by  :math:`(d)'`  
by multiplying the coordinate vector with a matrix :math:`H`. 
Multiplication with matrix :math:`H` is equivalent to
applying the Hadamard gate to Qubit 11 (i.e. to the bit with 
valence :math:`2^{11}`) of the coordinates of  :math:`v` .  
Thus :math:`H` is an involution. 

The numbering of the 4096 basis vectors :math:`(d)'` corresponds 
to the numbering of the positive elemments of :math:`\mathcal{P}`. 
We may put :math:`(d \oplus 2^{12})' = -(d)'`, where 
:math:`\oplus` means bitwise addition modulo 2; then that 
correspondence holds for all values :math:`0 \leq d < 2^{13}`. 
The exact defnition of the operator :math:`\oplus` on the Parker 
loop :math:`\mathcal{P}` is given in section  :ref:`implement-gen-mm`.

In this basis the operation of the extraspecial group 
:math:`Q_{x0}` is very simple:

    * :math:`x_e x_{\theta(e)}` maps  :math:`(d)'` to   
      :math:`(d \oplus e)'` for :math:`e \in \mathcal{P}`.
      Here :math:`\theta` is the cocycle, see section 
      :ref:`parker-loop-label`. In the language of 
      quantum computing this corresponds to a sequence
      of commuting **not** gates.
      
    * :math:`x_\epsilon` maps    :math:`(d)'` to   
      :math:`(-1)^{\langle d, \epsilon\rangle}(d)'`
      for :math:`\epsilon \in \mathcal{C}^*`. Here 
      :math:`\mathcal{C}^*` is the Golay cocode. In 
      the language of quantum computing this corresponds 
      to a sequence of commuting **phase**-:math:`\pi`
      gates.


Since :math:`H` is in :math:`\mathcal{C}_{12}`, it operates on 
:math:`Q_{x0}` by conjugation. The following subtlety has to be
considered in function ``conv_pauli_vector_xspecial`` (in 
file ``xsp2cco1.c``) implementing that operation. 
:math:`H` exchanges the anticcommuting
elements of :math:`Q_{x0}` of with numbers ``0x800`` and  
``0x800000``  in Leech lattice encoding. Thus it negates 
the element with number ``0x800800``.
 

Converting an element in **G_x0 representation** to a word of generators
........................................................................

Perhaps the most inportant function in this module is function
``xsp2co1_elem_to_word``. This function converts an element of 
the subgroup :math:`G_{x0}` from **G_x0 representation** to a 
word :math:`w` of generators of the monster group. That word 
:math:`w` is unique and **reduced** in the sense explained in the
description of that function.

There is also a companion function ``xsp2co1_reduce_word`` that
first converts an arbitrary word in the generators of the 
subgroup :math:`G_{x0}` of the monster to  **G_x0 representation**.
Then it calls function ``xsp2co1_elem_to_word`` to compute
the **reduced** word equal to the input word.

We briefly explain the implementation of function 
``xsp2co1_elem_to_word``. Let :math:`g \in G_{x0}` be given in
**G_x0 representation**. The function first computes the image
:math:`g^{-1} x_\Omega g` of :math:`x_\Omega` using function
``xsp2co1_xspecial_conjugate``, with :math:`\Omega` as in
:cite:`Seysen20`. Using that image and function 
``gen_leech2_reduce_type4`` in file ``gen_leech.c`` we can
compute a word :math:`w_1` in the generators of  :math:`G_{x0}` 
such that :math:`g_1 = g \cdot w_1` stabilizes  
:math:`x_\Omega` up to sign.

Then :math:`g_1` is in the monomial subgroup :math:`N_{x0}` of  
:math:`G_{x0}` of structure :math:`2^{1+24}.2^{11}.M_{24}`. For any 
:math:`g_1 \in N_{x0}` function ``xsp2co1_elem_monomial_to_xsp`` 
computes a word in the generators of the monster that is equal 
to :math:`g_1`. 

Given a monomial element :math:`g_1` in **G_x0 representation**,
function ``xsp2co1_elem_monomial_to_xsp`` computes a word of
generators equal to  :math:`g_1`  as follows.
Ignoring signs and identifying the basis vector :math:`d_1^+` of
:math:`4096_x` with the basis vector :math:`d_1^-`, Table 3 in
:cite:`Seysen20` states that :math:`x_\pi y_e x_f` maps  
:math:`d_1^\pm` to  :math:`(d^\pi e f)_1^\pm`. This corresponds
to an affine mapping :math:`d \mapsto d^\pi e f` from
:math:`\mathcal{C} / \langle \Omega \rangle` to itself, from which
we can easily compute :math:`e, f` (modulo sign and :math:`\Omega`),
and also :math:`\pi`. Thus we can find a word :math:`w_2` in the 
generators :math:`x_\pi y_e x_f`, such that :math:`g_2 = g_1 w_2`
stabilizes all basis vectors of :math:`4096_x` up to sign,
and possibly exchanging :math:`d_1^+` with :math:`d_1^-`.
Then :math:`g_2` can easily be converted into a word in 
the generators 
:math:`x_{-1}, x_\Omega, x_\delta, \delta \in \mathcal{C}^*`.
 
We use function ``qstate12_monomial_matrix_row_op`` in file 
``qmatrix12.c`` for obtaining the affine mapping 
:math:`d \mapsto d^\pi e f` from the **G_x0 representation**
of :math:`g_1`. We use the functions in file ``mat24_functions.c``
for obtaining :math:`\pi` from that affine mapping.

"""
